#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import json
import os
import sys
from datetime import datetime

import click
import yaml

from servicecatalog_puppet import (
    config,
    constants,
    environmental_variables,
    remote_config,
)
from servicecatalog_puppet.commands import (
    bootstrap as bootstrap_commands,
    graph as graph_commands,
    management as management_commands,
    manifest as manifest_commands,
    misc as misc_commands,
    orgs as orgs_commands,
    show_codebuilds as show_codebuilds_commands,
    show_pipelines as show_pipelines_commands,
    spoke_management as spoke_management_commands,
    task_reference as task_reference_commands,
    version as version_commands,
)


@click.group()
@click.option("--info/--no-info", default=False)
@click.option("--info-line-numbers/--no-info-line-numbers", default=False)
def cli(info, info_line_numbers):
    """cli for pipeline tools"""
    misc_commands.cli(info, info_line_numbers)


@cli.command()
@click.argument("f", type=click.File())
def graph(f):
    click.echo(graph_commands.graph(f))


@cli.command()
@click.argument("f", type=click.File())
@click.option("--single-account", default=None)
@click.option("--puppet-account-id", default=None)
def dry_run(f, single_account, puppet_account_id):
    click.echo("This command is disabled for now, due to performance issues.")


def parse_tags(ctx, param, value):
    tags = list()
    for val in value:
        k, v = val.split("=")
        tags.append(dict(Key=k, Value=v))
    return tags


@cli.command()
@click.argument("puppet_account_id")
@click.argument("iam_role_arns", nargs=-1)
@click.option(
    "--permission-boundary", default="arn:aws:iam::aws:policy/AdministratorAccess"
)
@click.option("--puppet-role-name", default="PuppetRole")
@click.option("--puppet-role-path", default="/servicecatalog-puppet/")
@click.option("--tag", multiple=True, callback=parse_tags, default=[])
def bootstrap_spoke_as(
    puppet_account_id,
    iam_role_arns,
    permission_boundary,
    puppet_role_name,
    puppet_role_path,
    tag,
):
    spoke_management_commands.bootstrap_spoke_as(
        puppet_account_id,
        iam_role_arns,
        permission_boundary,
        puppet_role_name,
        puppet_role_path,
        tag,
    )


@cli.command()
@click.argument("puppet_account_id")
@click.option(
    "--permission-boundary", default="arn:aws:iam::aws:policy/AdministratorAccess"
)
@click.option("--puppet-role-name", default="PuppetRole")
@click.option("--puppet-role-path", default="/servicecatalog-puppet/")
@click.option("--tag", multiple=True, callback=parse_tags, default=[])
def bootstrap_spoke(
    puppet_account_id, permission_boundary, puppet_role_name, puppet_role_path, tag
):
    spoke_management_commands.bootstrap_spoke(
        puppet_account_id, permission_boundary, puppet_role_name, puppet_role_path, tag
    )


@cli.command()
@click.argument("ou_path_or_id")
@click.argument("role_name")
@click.argument("iam_role_arns", nargs=-1)
@click.option(
    "--permission-boundary", default="arn:aws:iam::aws:policy/AdministratorAccess"
)
@click.option("--num-workers", default=10)
@click.option("--puppet-role-name", default="PuppetRole")
@click.option("--puppet-role-path", default="/servicecatalog-puppet/")
@click.option("--tag", multiple=True, callback=parse_tags, default=[])
def bootstrap_spokes_in_ou(
    ou_path_or_id,
    role_name,
    iam_role_arns,
    permission_boundary,
    num_workers,
    puppet_role_name,
    puppet_role_path,
    tag,
):
    setup_config()
    spoke_management_commands.bootstrap_spokes_in_ou(
        ou_path_or_id,
        role_name,
        iam_role_arns,
        permission_boundary,
        num_workers,
        puppet_role_name,
        puppet_role_path,
        tag,
    )


@cli.command()
@click.option("--with-manual-approvals/--with-no-manual-approvals", default=False)
@click.option(
    "--puppet-code-pipeline-role-permission-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
    envvar="PUPPET_CODE_PIPELINE_ROLE_PERMISSION_BOUNDARY",
)
@click.option(
    "--source-role-permissions-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
    envvar="SOURCE_ROLE_PERMISSIONS_BOUNDARY",
)
@click.option(
    "--puppet-generate-role-permission-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
    envvar="PUPPET_GENERATE_ROLE_PERMISSION_BOUNDARY",
)
@click.option(
    "--puppet-deploy-role-permission-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
    envvar="PUPPET_DEPLOY_ROLE_PERMISSION_BOUNDARY",
)
@click.option(
    "--puppet-provisioning-role-permissions-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
    envvar="PUPPET_PROVISIONING_ROLE_PERMISSIONS_BOUNDARY",
)
@click.option(
    "--cloud-formation-deploy-role-permissions-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
    envvar="CLOUD_FORMATION_DEPLOY_ROLE_PERMISSIONS_BOUNDARY",
)
@click.option(
    "--deploy_environment_compute_type",
    default="BUILD_GENERAL1_SMALL",
    show_default=True,
    envvar="DEPLOY_ENVIRONMENT_COMPUTE_TYPE",
)
@click.option(
    "--spoke_deploy_environment_compute_type",
    default="BUILD_GENERAL1_SMALL",
    show_default=True,
    envvar="SPOKE_DEPLOY_ENVIRONMENT_COMPUTE_TYPE",
)
@click.option(
    "--deploy_num_workers",
    default=10,
    type=click.INT,
    show_default=True,
    envvar="DEPLOY_NUM_WORKERS",
)
@click.option("--source-provider", default="CodeCommit", envvar="SCM_SOURCE_PROVIDER")
@click.option(
    "--repository_name", default="ServiceCatalogPuppet", envvar="SCM_REPOSITORY_NAME"
)
@click.option("--branch-name", default="master", envvar="SCM_BRANCH_NAME")
@click.option("--owner")
@click.option("--repo")
@click.option("--branch")
@click.option(
    "--poll-for-source-changes/--no-poll-for-source-changes",
    default=True,
    envvar="SCT_POLL_FOR_SOURCE_CHANGES",
)
@click.option("--webhook-secret")
@click.option("--puppet-role-name", default="PuppetRole", envvar="PUPPET_ROLE_NAME")
@click.option(
    "--puppet-role-path", default="/servicecatalog-puppet/", envvar="PUPPET_ROLE_PATH"
)
@click.option("--scm-connection-arn", envvar="SCM_CONNECTION_ARN")
@click.option(
    "--scm-full-repository-id",
    default="ServiceCatalogFactory",
    envvar="SCM_FULL_REPOSITORY_ID",
)
@click.option("--scm-branch-name", default="main", envvar="SCM_BRANCH_NAME")
@click.option("--scm-bucket-name", envvar="SCM_BUCKET_NAME")
@click.option(
    "--scm-object-key", default="ServiceCatalogPuppet.zip", envvar="SCM_OBJECT_KEY"
)
@click.option(
    "--create-repo/--no-create-repo", default=False, envvar="SCM_SHOULD_CREATE_REPO"
)
@click.option(
    "--should-validate/--no-should-validate",
    default=False,
    envvar="SCT_SHOULD_VALIDATE",
)
@click.option(
    "--custom-source-action-git-url",
    envvar="SCM_CUSTOM_SOURCE_ACTION_GIT_URL",
)
@click.option(
    "--custom-source-action-git-web-hook-ip-address",
    default="0.0.0.0/0",
    envvar="SCM_CUSTOM_SOURCE_ACTION_GIT_WEB_HOOK_IP_ADDRESS",
)
@click.option(
    "--custom-source-action-custom-action-type-version",
    default="CustomVersion1",
    envvar="SCM_CUSTOM_SOURCE_ACTION_CUSTOM_ACTION_TYPE_VERSION",
)
@click.option(
    "--custom-source-action-custom-action-type-provider",
    default="CustomProvider1",
    envvar="SCM_CUSTOM_SOURCE_ACTION_CUSTOM_ACTION_TYPE_PROVIDER",
)
def bootstrap(
    with_manual_approvals,
    puppet_code_pipeline_role_permission_boundary,
    source_role_permissions_boundary,
    puppet_generate_role_permission_boundary,
    puppet_deploy_role_permission_boundary,
    puppet_provisioning_role_permissions_boundary,
    cloud_formation_deploy_role_permissions_boundary,
    deploy_environment_compute_type,
    spoke_deploy_environment_compute_type,
    deploy_num_workers,
    source_provider,
    repository_name,
    branch_name,
    owner,
    repo,
    branch,
    poll_for_source_changes,
    webhook_secret,
    puppet_role_name,
    puppet_role_path,
    scm_connection_arn,
    scm_full_repository_id,
    scm_branch_name,
    scm_bucket_name,
    scm_object_key,
    create_repo,
    should_validate,
    custom_source_action_git_url,
    custom_source_action_git_web_hook_ip_address,
    custom_source_action_custom_action_type_version,
    custom_source_action_custom_action_type_provider,
):
    setup_config()
    puppet_account_id = config.get_puppet_account_id()

    parameters = dict(
        with_manual_approvals=with_manual_approvals,
        puppet_account_id=puppet_account_id,
        puppet_code_pipeline_role_permission_boundary=puppet_code_pipeline_role_permission_boundary,
        source_role_permissions_boundary=source_role_permissions_boundary,
        puppet_generate_role_permission_boundary=puppet_generate_role_permission_boundary,
        puppet_deploy_role_permission_boundary=puppet_deploy_role_permission_boundary,
        puppet_provisioning_role_permissions_boundary=puppet_provisioning_role_permissions_boundary,
        cloud_formation_deploy_role_permissions_boundary=cloud_formation_deploy_role_permissions_boundary,
        deploy_environment_compute_type=deploy_environment_compute_type,
        spoke_deploy_environment_compute_type=spoke_deploy_environment_compute_type,
        deploy_num_workers=deploy_num_workers,
        source_provider=source_provider,
        owner=None,
        repo=None,
        branch=None,
        poll_for_source_changes=poll_for_source_changes,
        webhook_secret=webhook_secret,
        puppet_role_name=puppet_role_name,
        puppet_role_path=puppet_role_path,
        scm_connection_arn=None,
        scm_full_repository_id=None,
        scm_branch_name=None,
        scm_bucket_name=None,
        scm_object_key=None,
        scm_skip_creation_of_repo=not create_repo,
        should_validate=should_validate,
        custom_source_action_git_url=None,
        custom_source_action_git_web_hook_ip_address=None,
        custom_source_action_custom_action_type_version=None,
        custom_source_action_custom_action_type_provider=None,
    )
    if source_provider == "CodeCommit":
        parameters.update(
            dict(
                repo=repository_name,
                branch=branch_name,
            )
        )
    elif source_provider == "GitHub":
        parameters.update(
            dict(
                owner=owner,
                repo=repo,
                branch=branch,
                webhook_secret=webhook_secret,
            )
        )
    elif source_provider == "CodeStarSourceConnection":
        parameters.update(
            dict(
                scm_connection_arn=scm_connection_arn,
                scm_full_repository_id=scm_full_repository_id,
                scm_branch_name=scm_branch_name,
            )
        )
    elif source_provider == "S3":
        parameters.update(
            dict(
                scm_bucket_name=scm_bucket_name,
                scm_object_key=scm_object_key,
            )
        )
    elif source_provider == "Custom":
        parameters.update(
            dict(
                custom_source_action_git_url=custom_source_action_git_url,
                branch=branch_name,
                custom_source_action_git_web_hook_ip_address=custom_source_action_git_web_hook_ip_address,
                custom_source_action_custom_action_type_version=custom_source_action_custom_action_type_version,
                custom_source_action_custom_action_type_provider=custom_source_action_custom_action_type_provider,
            )
        )
    else:
        raise Exception(f"Unsupported source provider: {source_provider}")

    bootstrap_commands.bootstrap(**parameters)


@cli.command()
@click.argument("complexity", default="simple")
@click.argument("p", type=click.Path(exists=True))
def seed(complexity, p):
    management_commands.seed(complexity, p)


@cli.command()
@click.argument("expanded_manifest", type=click.File())
@click.option("--format", "-f", type=click.Choice(["table", "json"]), default="table")
def list_launches(expanded_manifest, format):
    click.echo("This command is disabled for now, due to performance issues.")


@cli.command()
@click.argument("f", type=click.File())
@click.option(
    "--single-account",
    default="None",
    envvar="SINGLE_ACCOUNT",
)
@click.option(
    "--single-action-section",
    default="*",
    envvar="SINGLE_ACTION_SECTION",
)
@click.option(
    "--single-action-item",
    default="*",
    envvar="SINGLE_ACTION_ITEM",
)
@click.option(
    "--single-action-include-dependencies",
    default="Yes",
    envvar="SINGLE_ACTION_INCLUDE_DEPENDENCIES",
)
@click.option(
    "--single-action-include-reverse-dependencies",
    default="Yes",
    envvar="SINGLE_ACTION_INCLUDE_REVERSE_DEPENDENCIES",
)
def expand(
    f,
    single_account,
    single_action_section,
    single_action_item,
    single_action_include_dependencies,
    single_action_include_reverse_dependencies,
):
    puppet_account_id = remote_config.get_puppet_account_id()
    regions = remote_config.get_regions(puppet_account_id, constants.HOME_REGION)
    overrides = dict()
    if single_account != "" and single_account != "None":
        overrides["single_account"] = single_account
    if single_action_section != "*":
        overrides["single_action_section"] = single_action_section
    if single_action_item != "*":
        overrides["single_action_item"] = single_action_item
    overrides["single_action_include_dependencies"] = (
        str(single_action_include_dependencies).lower() == "yes"
    )
    overrides["single_action_include_reverse_dependencies"] = (
        str(single_action_include_reverse_dependencies).lower() == "yes"
    )

    print(f"overrides is {overrides}")

    if overrides.get("subset"):
        subset = dict(**overrides.get("subset"))
    else:
        subset = dict(**overrides)

    if subset.get("single_account"):
        del subset["single_account"]

    if overrides.get("single_account"):
        del overrides["single_account"]

    extra_params = dict(single_account=overrides.get("single_account"), subset=subset)
    click.echo(f"Overridden parameters {extra_params}")

    manifest_commands.expand(f, puppet_account_id, regions, **extra_params)
    if config.get_should_explode_manifest(puppet_account_id):
        manifest_commands.explode(f)


@cli.command()
@click.argument("f", type=click.File())
def generate_task_reference(
    f,
):
    setup_config()
    task_reference_commands.generate_task_reference(f)


def setup_config(
    puppet_account_id=None,
    single_account=None,
    execution_mode="hub",
    num_workers="10",
    home_region=None,
    regions="",
    should_collect_cloudformation_events="None",
    should_forward_events_to_eventbridge="None",
    should_forward_failures_to_opscenter="None",
    output_cache_starting_point="",
    is_caching_enabled="",
    global_sharing_mode_default="",
    global_share_tag_options_default="",
    global_share_principals_default="",
    on_complete_url=None,
):
    home_region_to_use = home_region or constants.HOME_REGION
    if puppet_account_id is None:
        puppet_account_id_to_use = remote_config.get_puppet_account_id()
    else:
        puppet_account_id_to_use = puppet_account_id
    os.environ[environmental_variables.PUPPET_ACCOUNT_ID] = puppet_account_id_to_use
    print(f"single_account is {single_account}")
    if single_account:
        print("I AM SETTING single_account uh oh")
        os.environ[environmental_variables.SINGLE_ACCOUNT_ID] = single_account

    os.environ[environmental_variables.EXECUTION_MODE] = execution_mode
    os.environ[environmental_variables.NUM_WORKERS] = num_workers
    os.environ[environmental_variables.HOME_REGION] = home_region_to_use
    os.environ[environmental_variables.EXECUTOR_ACCOUNT_ID] = str(
        remote_config.get_current_account_id()
    )
    os.environ[environmental_variables.REGIONS] = json.dumps(
        regions
        if regions != ""
        else remote_config.get_regions(puppet_account_id_to_use, home_region_to_use)
    )
    if os.environ.get(environmental_variables.DRIFT_TOKEN):
        click.echo(
            f"Found existing {environmental_variables.DRIFT_TOKEN}: {os.environ.get(environmental_variables.DRIFT_TOKEN)}"
        )
    else:
        drift_token = remote_config.get_drift_token(
            puppet_account_id_to_use, home_region_to_use
        )
        if drift_token is None:
            drift_token = str(datetime.now())
            click.echo(f"Created new drift_token: {drift_token}")
        else:
            click.echo(f"Found existing drift_token: {drift_token}")
        os.environ[environmental_variables.DRIFT_TOKEN] = drift_token

    run_token = os.environ.get(environmental_variables.RUN_TOKEN)
    if run_token:
        click.echo(f"Found existing run_token: {run_token}")
    else:
        run_token = str(datetime.now())
        os.environ[environmental_variables.RUN_TOKEN] = run_token
        click.echo(f"Created new run_token: {run_token}")

    os.environ[environmental_variables.SHOULD_USE_SNS] = (
        str(remote_config.get_should_use_sns(puppet_account_id_to_use, home_region))
        if should_collect_cloudformation_events == "None"
        else str(should_collect_cloudformation_events)
    )

    os.environ[environmental_variables.SHOULD_FORWARD_EVENTS_TO_EVENTBRIDGE] = str(
        str(
            remote_config.get_should_use_eventbridge(
                puppet_account_id_to_use, home_region
            )
        )
        if should_forward_events_to_eventbridge == "None"
        else should_forward_events_to_eventbridge
    )
    os.environ[environmental_variables.SHOULD_FORWARD_FAILURES_TO_OPSCENTER] = str(
        str(
            remote_config.get_should_forward_failures_to_opscenter(
                puppet_account_id_to_use, home_region
            )
        )
        if should_forward_failures_to_opscenter == "None"
        else should_forward_failures_to_opscenter
    )

    if not os.environ.get(
        environmental_variables.SHOULD_DELETE_ROLLBACK_COMPLETE_STACKS
    ):
        os.environ[environmental_variables.SHOULD_DELETE_ROLLBACK_COMPLETE_STACKS] = (
            str(
                remote_config.get_should_delete_rollback_complete_stacks(
                    puppet_account_id_to_use
                )
            )
        )

    if not os.environ.get(environmental_variables.SHOULD_USE_PRODUCT_PLANS):
        os.environ[environmental_variables.SHOULD_USE_PRODUCT_PLANS] = str(
            remote_config.get_should_use_product_plans(
                puppet_account_id_to_use, os.environ.get("AWS_DEFAULT_REGION")
            )
        )

    if not os.environ.get(environmental_variables.INITIALISER_STACK_TAGS):
        os.environ[environmental_variables.INITIALISER_STACK_TAGS] = json.dumps(
            remote_config.get_initialiser_stack_tags()
        )  # TODO make dynamic

    os.environ[environmental_variables.VERSION] = constants.VERSION
    os.environ[environmental_variables.OUTPUT_CACHE_STARTING_POINT] = (
        output_cache_starting_point
    )

    if is_caching_enabled == "":
        os.environ[environmental_variables.IS_CACHING_ENABLED] = str(
            remote_config.is_caching_enabled(puppet_account_id_to_use, home_region)
        )
    else:
        os.environ[environmental_variables.IS_CACHING_ENABLED] = str(is_caching_enabled)

    if global_sharing_mode_default == "":
        os.environ[environmental_variables.GLOBAL_SHARING_MODE] = (
            remote_config.get_global_sharing_mode_default(
                puppet_account_id_to_use, home_region
            )
        )
    else:
        os.environ[environmental_variables.GLOBAL_SHARING_MODE] = str(
            global_sharing_mode_default
        )

    if global_share_tag_options_default == "":
        os.environ[environmental_variables.GLOBAL_SHARE_TAG_OPTIONS] = str(
            remote_config.get_global_share_tag_options_default(
                puppet_account_id_to_use, home_region
            )
        )
    else:
        os.environ[environmental_variables.GLOBAL_SHARE_TAG_OPTIONS] = (
            global_share_tag_options_default
        )

    if global_share_principals_default == "":
        os.environ[environmental_variables.GLOBAL_SHARE_PRINCIPALS] = str(
            remote_config.get_global_share_principals_default(
                puppet_account_id_to_use, home_region
            )
        )
    else:
        os.environ[environmental_variables.GLOBAL_SHARE_PRINCIPALS] = (
            global_share_principals_default
        )

    if on_complete_url:
        os.environ[environmental_variables.ON_COMPLETE_URL] = on_complete_url
    if not os.environ.get(environmental_variables.SPOKE_EXECUTION_MODE_DEPLOY_ENV):
        os.environ[environmental_variables.SPOKE_EXECUTION_MODE_DEPLOY_ENV] = (
            remote_config.get_spoke_deploy_environment_compute_type(
                puppet_account_id_to_use, home_region
            )
        )
    if not os.environ.get(environmental_variables.SCHEDULER_THREADS_OR_PROCESSES):
        os.environ[environmental_variables.SCHEDULER_THREADS_OR_PROCESSES] = (
            remote_config.get_scheduler_threads_or_processes(
                puppet_account_id_to_use, home_region
            )
        )
    if not os.environ.get(environmental_variables.SCHEDULER_ALGORITHM):
        os.environ[environmental_variables.SCHEDULER_ALGORITHM] = (
            remote_config.get_scheduler_algorithm(puppet_account_id_to_use, home_region)
        )
    if not os.environ.get(environmental_variables.AWS_STS_REGIONAL_ENDPOINTS):
        os.environ[environmental_variables.AWS_STS_REGIONAL_ENDPOINTS] = (
            remote_config.get_aws_sts_regional_endpoints(
                puppet_account_id_to_use, home_region
            )
        )


@cli.command()
@click.argument("path", type=click.Path(exists=True))
@click.option("--num-workers", default=10)
@click.option("--execution-mode", default="hub")
@click.option("--puppet-account-id", default=None)
@click.option("--home-region", default=None)
@click.option("--regions", default="")
@click.option("--should-collect-cloudformation-events", default=None, type=bool)
@click.option("--should-forward-events-to-eventbridge", default=None, type=bool)
@click.option("--should-forward-failures-to-opscenter", default=None, type=bool)
@click.option(
    "--output-cache-starting-point",
    default="",
    show_default=True,
    envvar="OUTPUT_CACHE_STARTING_POINT",
)
@click.option(
    "--is-caching-enabled",
    default="",
    envvar="SCT_IS_CACHING_ENABLED",
)
@click.option(
    "--single-account",
    default="None",
    envvar="SINGLE_ACCOUNT",
)
@click.option(
    "--single-action-section",
    default="*",
    envvar="SINGLE_ACTION_SECTION",
)
@click.option(
    "--single-action-item",
    default="*",
    envvar="SINGLE_ACTION_ITEM",
)
@click.option(
    "--single-action-include-dependencies",
    default="Yes",
    envvar="SINGLE_ACTION_INCLUDE_DEPENDENCIES",
)
@click.option(
    "--single-action-include-reverse-dependencies",
    default="Yes",
    envvar="SINGLE_ACTION_INCLUDE_REVERSE_DEPENDENCIES",
)
@click.option("--on-complete-url", default=None)
def deploy_from_task_reference(
    path,
    num_workers,
    execution_mode,
    puppet_account_id,
    home_region,
    regions,
    should_collect_cloudformation_events,
    should_forward_events_to_eventbridge,
    should_forward_failures_to_opscenter,
    output_cache_starting_point,
    is_caching_enabled,
    single_account,
    single_action_section,
    single_action_item,
    single_action_include_dependencies,
    single_action_include_reverse_dependencies,
    on_complete_url,
):
    params = dict()

    overrides = dict()
    if single_account != "" and single_account != "None":
        overrides["single_account"] = single_account
    if single_action_section != "*":
        overrides["single_action_section"] = single_action_section
    if single_action_item != "*":
        overrides["single_action_item"] = single_action_item
    overrides["single_action_include_dependencies"] = (
        str(single_action_include_dependencies).lower() == "yes"
    )
    overrides["single_action_include_reverse_dependencies"] = (
        str(single_action_include_reverse_dependencies).lower() == "yes"
    )

    if overrides.get("subset"):
        subset = overrides.get("subset")
        overrides = dict(
            section=subset.get("section"),
            item=subset.get("name"),
            include_dependencies=subset.get("include_dependencies"),
            include_reverse_dependencies=subset.get("include_reverse_dependencies"),
        )
    single_account = (
        str(overrides.get("single_account"))
        if overrides.get("single_account")
        else None
    )
    params.update(
        dict(
            single_account=single_account,
            subset=overrides,
        )
    )
    click.echo(f"Overridden parameters {params}")

    setup_config(
        puppet_account_id=puppet_account_id,
        single_account=params.get("single_account"),
        num_workers=str(num_workers),
        execution_mode=execution_mode,
        home_region=home_region,
        regions=regions,
        should_collect_cloudformation_events=str(should_collect_cloudformation_events),
        should_forward_events_to_eventbridge=str(should_forward_events_to_eventbridge),
        should_forward_failures_to_opscenter=str(should_forward_failures_to_opscenter),
        output_cache_starting_point=output_cache_starting_point,
        is_caching_enabled=is_caching_enabled,
        on_complete_url=on_complete_url,
    )

    click.echo(
        f"running in partition: {config.get_partition()} as {config.get_puppet_role_path()}{config.get_puppet_role_name()}"
    )

    task_reference_commands.deploy_from_task_reference(path)


@cli.command()
@click.argument("p", type=click.Path())
@click.option("--num-workers", default=10)
@click.option("--execution-mode", default="hub")
@click.option("--puppet-account-id")
@click.option("--single-account", default=None)
@click.option("--home-region", default=None)
@click.option("--regions", default="")
@click.option(
    "--should-collect-cloudformation-events",
    default=None,
    type=bool,
    envvar="SHOULD_COLLECT_CLOUDFORMATION_EVENTS",
)
@click.option(
    "--should-forward-events-to-eventbridge",
    default=None,
    type=bool,
    envvar="SHOULD_FORWARD_EVENTS_TO_EVENTBRIDGE",
)
@click.option(
    "--should-forward-failures-to-opscenter",
    default=None,
    type=bool,
    envvar="SHOULD_FORWARD_FAILURES_TO_OPSCENTER",
)
@click.option(
    "--output-cache-starting-point",
    default="",
    show_default=True,
    envvar="OUTPUT_CACHE_STARTING_POINT",
)
@click.option(
    "--is-caching-enabled",
    default="",
    envvar="SCT_IS_CACHING_ENABLED",
)
@click.option(
    "--global-sharing-mode-default",
    default="",
    envvar="SCT_GLOBAL_SHARING_MODE",
)
@click.option(
    "--global-share-tag-options-default",
    default="",
    envvar="SCT_GLOBAL_SHARE_TAG_OPTIONS",
)
@click.option(
    "--global-share-principals-default",
    default="",
    envvar="SCT_GLOBAL_SHARE_PRINCIPALS",
)
def deploy_in_spoke_from_task_reference(
    p,
    num_workers,
    execution_mode,
    puppet_account_id,
    single_account,
    home_region,
    regions,
    should_collect_cloudformation_events,
    should_forward_events_to_eventbridge,
    should_forward_failures_to_opscenter,
    output_cache_starting_point,
    is_caching_enabled,
    global_sharing_mode_default,
    global_share_tag_options_default,
    global_share_principals_default,
):
    setup_config(
        puppet_account_id=puppet_account_id,
        single_account=single_account,
        num_workers=str(num_workers),
        execution_mode=execution_mode,
        home_region=home_region,
        regions=regions,
        should_collect_cloudformation_events=str(should_collect_cloudformation_events),
        should_forward_events_to_eventbridge=str(should_forward_events_to_eventbridge),
        should_forward_failures_to_opscenter=str(should_forward_failures_to_opscenter),
        output_cache_starting_point=output_cache_starting_point,
        is_caching_enabled=is_caching_enabled,
        global_sharing_mode_default=global_sharing_mode_default,
        global_share_tag_options_default=global_share_tag_options_default,
        global_share_principals_default=global_share_principals_default,
    )
    click.echo(
        f"running in partition: {config.get_partition()} as {config.get_puppet_role_path()}{config.get_puppet_role_name()}"
    )

    task_reference_commands.deploy_from_task_reference(p)


@cli.command()
@click.argument("f", type=click.File())
def validate(f):
    setup_config()
    manifest_commands.validate(f)


@cli.command()
def version():
    setup_config()
    version_commands.version()


@cli.command()
@click.option(
    "--filter", type=click.Choice(["none", "single-runs", "full-runs"]), default="none"
)
@click.option("--format", type=click.Choice(["csv", "json"]), default="csv")
@click.option("--limit", type=click.INT, default=20)
def show_codebuilds(filter, format, limit):
    setup_config()
    show_codebuilds_commands.show_codebuilds(filter, limit, format)


@cli.command()
@click.argument("p", type=click.Path(exists=True))
def upload_config(p):
    setup_config()
    content = open(p, "r").read()
    management_commands.upload_config(yaml.safe_load(content))


@cli.command()
@click.argument("org-iam-role-arn")
def set_org_iam_role_arn(org_iam_role_arn):
    setup_config()
    orgs_commands.set_org_iam_role_arn(org_iam_role_arn)


@cli.command()
@click.argument("org-iam-role-arn")
def set_org_scp_role_arn(org_iam_role_arn):
    setup_config()
    orgs_commands.set_org_scp_role_arn(org_iam_role_arn)


@cli.command()
@click.argument("puppet_account_id")
@click.option("--tag", multiple=True, callback=parse_tags, default=[])
def bootstrap_org_master(puppet_account_id, tag):
    setup_config()
    orgs_commands.bootstrap_org_master(puppet_account_id, tag)


@cli.command()
@click.argument("puppet_account_id")
@click.option("--tag", multiple=True, callback=parse_tags, default=[])
def bootstrap_scp_master(puppet_account_id, tag):
    setup_config()
    orgs_commands.bootstrap_scp_master(puppet_account_id, tag)


@cli.command()
@click.argument("what", default="puppet")
@click.option("--tail/--no-tail", default=False)
def run(what, tail):
    setup_config()
    misc_commands.run(what, tail)


@cli.command()
def list_resources():
    setup_config()
    management_commands.list_resources()


@cli.command()
@click.argument("f", type=click.File())
@click.argument("name")
@click.argument("portfolio_name")
def import_product_set(f, name, portfolio_name):
    setup_config()
    manifest_commands.import_product_set(f, name, portfolio_name)


@cli.command()
@click.argument("account_or_ou_file_path", type=click.File())
def add_to_accounts(account_or_ou_file_path):
    setup_config()
    manifest_commands.add_to_accounts(yaml.safe_load(account_or_ou_file_path))


@cli.command()
@click.argument("account_id_or_ou_id_or_ou_path")
def remove_from_accounts(account_id_or_ou_id_or_ou_path):
    setup_config()
    manifest_commands.remove_from_accounts(account_id_or_ou_id_or_ou_path)


@cli.command()
@click.argument("launch_file_path", type=click.File())
def add_to_launches(launch_name, launch_file_path):
    setup_config()
    manifest_commands.add_to_launches(launch_name, yaml.safe_load(launch_file_path))


@cli.command()
@click.argument("launch_name")
def remove_from_launches(launch_name):
    setup_config()
    manifest_commands.remove_from_launches(launch_name)


@cli.command()
@click.argument("f", type=click.File())
def reset_provisioned_product_owner(f):
    raise Exception(
        "This is not supported within this version, please try an older version"
    )


@cli.command()
@click.argument("regions", nargs=-1)
def set_regions(regions):
    management_commands.set_config_value("regions", regions)


@cli.command()
@click.argument("name")
@click.argument("value")
def set_config_value(name, value):
    is_a_boolean = dict(
        spoke_deploy_environment_compute_type=False,
        scheduler_threads_or_processes=False,
        aws_sts_regional_endpoints=False,
    ).get(name, True)
    management_commands.set_config_value(name, value, is_a_boolean)


@cli.command()
@click.argument("name")
@click.argument("value")
def set_named_config_value(name, value):
    management_commands.set_named_config_value(name, value)


@cli.command()
@click.argument("execution_id")
@click.option("--puppet-account-id", default=None)
def export_puppet_pipeline_logs(execution_id, puppet_account_id):
    setup_config(puppet_account_id=puppet_account_id)
    if puppet_account_id is None:
        puppet_account_id = config.get_puppet_account_id()
    management_commands.export_puppet_pipeline_logs(execution_id, puppet_account_id)


@cli.command()
@click.argument("codebuild_execution_id")
@click.option("--puppet-account-id", default=None)
@click.option("--group-by-pid/--no-group-by-pid", default=False)
def generate_deploy_viz(codebuild_execution_id, puppet_account_id, group_by_pid):
    setup_config(puppet_account_id=puppet_account_id)
    if puppet_account_id is None:
        puppet_account_id = config.get_puppet_account_id()
    management_commands.export_deploy_viz(
        codebuild_execution_id, group_by_pid, puppet_account_id
    )


@cli.command()
@click.argument("codebuild_execution_id")
@click.option("--puppet-account-id", default=None)
def export_traces(codebuild_execution_id, puppet_account_id):
    setup_config(puppet_account_id=puppet_account_id)
    if puppet_account_id is None:
        puppet_account_id = config.get_puppet_account_id()
    management_commands.export_traces(codebuild_execution_id, puppet_account_id)


@cli.command()
@click.option("--puppet-account-id", default=None)
def uninstall(puppet_account_id):
    setup_config(puppet_account_id=puppet_account_id)
    if puppet_account_id is None:
        puppet_account_id = config.get_puppet_account_id()
    misc_commands.uninstall(puppet_account_id)


@cli.command()
@click.option("--puppet-account-id", default=None)
def release_spoke(puppet_account_id):
    setup_config(puppet_account_id=puppet_account_id)
    if puppet_account_id is None:
        puppet_account_id = config.get_puppet_account_id()
    spoke_management_commands.release_spoke(puppet_account_id)


@cli.command()
@click.argument("iam_role_arns", nargs=-1)
def wait_for_code_build_in(iam_role_arns):
    misc_commands.wait_for_code_build_in(iam_role_arns)
    misc_commands.wait_for_cloudformation_in(iam_role_arns)
    click.echo("AWS CodeBuild is available")


@cli.command()
@click.argument("execution")
@click.option("--on-complete-url", default=None)
def wait_for_run_to_complete(execution, on_complete_url):
    setup_config(puppet_account_id=remote_config.get_puppet_account_id())
    click.echo("Starting to wait for parameterised run to complete")
    succeeded = misc_commands.wait_for_run_to_complete(execution, on_complete_url)
    if succeeded:
        click.echo(f"Finished: 'SUCCESS'")
        sys.exit(0)
    else:
        click.echo(f"Finished: 'FAILED'")
        sys.exit(1)


@cli.command()
@click.option(
    "--format", "-f", type=click.Choice(["table", "json", "html"]), default="table"
)
def show_pipelines(format):
    show_pipelines_commands.show_pipelines(format)


if __name__ == "__main__":
    cli()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import os

from betterboto import client as betterboto_client
from botocore.config import Config

from servicecatalog_puppet import config, constants


class ClientMixin:
    def is_running_in_spoke(self):
        return self.execution_mode == constants.EXECUTION_MODE_SPOKE

    def spoke_client(self, service):
        kwargs = dict()
        if os.environ.get(f"CUSTOM_ENDPOINT_{service}"):
            kwargs["endpoint_url"] = os.environ.get(f"CUSTOM_ENDPOINT_{service}")
        return betterboto_client.CrossAccountClientContextManager(
            service,
            config.get_puppet_role_arn(self.account_id),
            f"{self.account_id}-{config.get_puppet_role_name()}",
            **kwargs,
        )

    def cross_account_client(self, account_id, service, region_name=None):
        region = region_name or self.region
        kwargs = dict(region_name=region)
        if os.environ.get(f"CUSTOM_ENDPOINT_{service}"):
            kwargs["endpoint_url"] = os.environ.get(f"CUSTOM_ENDPOINT_{service}")

        return betterboto_client.CrossAccountClientContextManager(
            service,
            config.get_puppet_role_arn(account_id),
            f"{account_id}-{region}-{config.get_puppet_role_name()}",
            **kwargs,
        )

    def spoke_regional_client(self, service, region_name=None, retry_max_attempts=None):
        region = region_name or self.region
        kwargs = dict(region_name=region)
        if retry_max_attempts is not None:
            kwargs["config"] = Config(
                retries={
                    "max_attempts": retry_max_attempts,
                }
            )
        if os.environ.get(f"CUSTOM_ENDPOINT_{service}"):
            kwargs["endpoint_url"] = os.environ.get(f"CUSTOM_ENDPOINT_{service}")

        return betterboto_client.CrossAccountClientContextManager(
            service,
            config.get_puppet_role_arn(self.account_id),
            f"{self.account_id}-{region}-{config.get_puppet_role_name()}",
            **kwargs,
        )

    def hub_client(self, service):
        kwargs = dict()
        if os.environ.get(f"CUSTOM_ENDPOINT_{service}"):
            kwargs["endpoint_url"] = os.environ.get(f"CUSTOM_ENDPOINT_{service}")
        if self.is_running_in_spoke():
            return betterboto_client.CrossAccountClientContextManager(
                service,
                config.get_puppet_role_arn(self.executor_account_id),
                f"{self.executor_account_id}-{config.get_puppet_role_name()}",
                **kwargs,
            )
        else:
            return betterboto_client.CrossAccountClientContextManager(
                service,
                config.get_puppet_role_arn(self.puppet_account_id),
                f"{self.puppet_account_id}-{config.get_puppet_role_name()}",
                **kwargs,
            )

    def hub_regional_client(self, service, region_name=None):
        region = region_name or self.region
        kwargs = dict(region_name=region)
        if os.environ.get(f"CUSTOM_ENDPOINT_{service}"):
            kwargs["endpoint_url"] = os.environ.get(f"CUSTOM_ENDPOINT_{service}")

        if self.is_running_in_spoke():
            return betterboto_client.CrossAccountClientContextManager(
                service,
                config.get_puppet_role_arn(self.executor_account_id),
                f"{self.executor_account_id}-{config.get_puppet_role_name()}",
                **kwargs,
            )
        else:
            return betterboto_client.CrossAccountClientContextManager(
                service,
                config.get_puppet_role_arn(self.puppet_account_id),
                f"{self.puppet_account_id}-{region}-{config.get_puppet_role_name()}",
                **kwargs,
            )

    def organizations_policy_client(self):
        kwargs = dict()
        if os.environ.get(f"CUSTOM_ENDPOINT_organizations"):
            kwargs["endpoint_url"] = os.environ.get(f"CUSTOM_ENDPOINT_organizations")
        if self.is_running_in_spoke():
            raise Exception("Cannot use organizations client in spoke execution")
        else:
            return betterboto_client.CrossAccountClientContextManager(
                "organizations",
                config.get_org_scp_role_arn(self.puppet_account_id),
                "org_scp_role_arn",
            )



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import troposphere as t
from troposphere import s3, sns, ssm


def get_template(version: str, default_region_value) -> t.Template:
    description = f"""Bootstrap template used to bootstrap a region of ServiceCatalog-Puppet master
{{"version": "{version}", "framework": "servicecatalog-puppet", "role": "bootstrap-master-region"}}"""

    template = t.Template(Description=description)

    version_parameter = template.add_parameter(
        t.Parameter("Version", Default=version, Type="String")
    )
    default_region_value_parameter = template.add_parameter(
        t.Parameter("DefaultRegionValue", Default=default_region_value, Type="String")
    )

    template.add_resource(
        ssm.Parameter(
            "DefaultRegionParam",
            Name="/servicecatalog-puppet/home-region",
            Type="String",
            Value=t.Ref(default_region_value_parameter),
            Tags={"ServiceCatalogPuppet:Actor": "Framework"},
        )
    )
    version_ssm_parameter = template.add_resource(
        ssm.Parameter(
            "Param",
            Name="service-catalog-puppet-regional-version",
            Type="String",
            Value=t.Ref(version_parameter),
            Tags={"ServiceCatalogPuppet:Actor": "Framework"},
        )
    )

    template.add_resource(
        s3.Bucket(
            "PipelineArtifactBucket",
            BucketName=t.Sub(
                "sc-puppet-pipeline-artifacts-${AWS::AccountId}-${AWS::Region}"
            ),
            VersioningConfiguration=s3.VersioningConfiguration(Status="Enabled"),
            BucketEncryption=s3.BucketEncryption(
                ServerSideEncryptionConfiguration=[
                    s3.ServerSideEncryptionRule(
                        ServerSideEncryptionByDefault=s3.ServerSideEncryptionByDefault(
                            SSEAlgorithm="AES256"
                        )
                    )
                ]
            ),
            PublicAccessBlockConfiguration=s3.PublicAccessBlockConfiguration(
                BlockPublicAcls=True,
                BlockPublicPolicy=True,
                IgnorePublicAcls=True,
                RestrictPublicBuckets=True,
            ),
            Tags=t.Tags({"ServiceCatalogPuppet:Actor": "Framework"}),
        )
    )

    regional_product_topic = template.add_resource(
        sns.Topic(
            "RegionalProductTopic",
            DisplayName="servicecatalog-puppet-cloudformation-regional-events",
            TopicName="servicecatalog-puppet-cloudformation-regional-events",
            Subscription=[
                sns.Subscription(
                    Endpoint=t.Sub(
                        "arn:${AWS::Partition}:sqs:${DefaultRegionValue}:${AWS::AccountId}:servicecatalog-puppet-cloudformation-events"
                    ),
                    Protocol="sqs",
                )
            ],
        ),
    )

    template.add_output(
        t.Output("Version", Value=t.GetAtt(version_ssm_parameter, "Value"))
    )
    template.add_output(
        t.Output("RegionalProductTopic", Value=t.Ref(regional_product_topic))
    )

    return template



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import io
import zipfile

import luigi

from servicecatalog_puppet import constants, serialisation_utils
from servicecatalog_puppet.workflow.dependencies import tasks


class TerminateWorkspaceTask(tasks.TaskWithParameters):
    workspace_name = luigi.Parameter()
    region = luigi.Parameter()
    account_id = luigi.Parameter()

    bucket = luigi.Parameter()
    key = luigi.Parameter()
    version_id = luigi.Parameter()

    ssm_param_inputs = luigi.ListParameter(default=[], significant=False)

    launch_parameters = luigi.DictParameter(default={}, significant=False)
    manifest_parameters = luigi.DictParameter(default={}, significant=False)
    account_parameters = luigi.DictParameter(default={}, significant=False)

    retry_count = luigi.IntParameter(default=1, significant=False)
    worker_timeout = luigi.IntParameter(default=0, significant=False)
    ssm_param_outputs = luigi.ListParameter(default=[], significant=False)
    requested_priority = luigi.IntParameter(significant=False, default=0)

    execution = luigi.Parameter()
    manifest_file_path = luigi.Parameter()

    section_name = constants.WORKSPACES
    cachable_level = constants.CACHE_LEVEL_RUN

    @property
    def item_name(self):
        return self.workspace_name

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "workspace_name": self.workspace_name,
            "region": self.region,
            "account_id": self.account_id,
        }

    def run(self):
        with self.hub_client("s3") as s3:
            options = (
                zipfile.ZipFile(
                    io.BytesIO(
                        s3.get_object(Bucket=self.bucket, Key=self.key)
                        .get("Body")
                        .read()
                    )
                )
                .open(f"options.json", "r")
                .read()
            )

        options = serialisation_utils.json_loads(options)

        zip_file_path = f"s3://{self.bucket}/{self.key}"
        state_file_path = f"s3://sc-puppet-state-{self.account_id}/workspace/{self.workspace_name}/{self.account_id}/{self.region}.zip"
        with self.spoke_client("codebuild") as codebuild:
            parameters_to_use = [
                dict(
                    name="TARGET_ACCOUNT",
                    value=self.account_id,
                    type="PLAINTEXT",
                ),
                dict(
                    name="STATE_FILE",
                    value=state_file_path,
                    type="PLAINTEXT",
                ),
                dict(
                    name="ZIP",
                    value=zip_file_path,
                    type="PLAINTEXT",
                ),
            ]

            for parameter_name, parameter_value in self.get_parameter_values().items():
                parameters_to_use.append(
                    dict(
                        name=f"TF_VAR_{parameter_name}",
                        value=f"{parameter_value}",
                        type="PLAINTEXT",
                    ),
                )

            parameters_to_use.append(
                dict(
                    name="TERRAFORM_VERSION",
                    value=options.get("Terraform", {}).get(
                        "Version", constants.DEFAULT_TERRAFORM_VERSION_VALUE
                    ),
                    type="PLAINTEXT",
                ),
            )

            codebuild.start_build_and_wait_for_completion(
                projectName=constants.TERMINATE_TERRAFORM_PROJECT_NAME,
                environmentVariablesOverride=parameters_to_use,
            )

        self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import logging

import luigi
from luigi import format
from luigi.contrib import s3

from servicecatalog_puppet import config, constants, serialisation_utils


logger = logging.getLogger(constants.PUPPET_LOGGER_NAME)


def escape(input):
    return "".join(x for x in input if x.isalnum() or x in "._- /")


class IOMixin:
    @property
    def drift_token_parameters(self):
        raise NotImplementedError(
            f"{self.__class__.__name__} has cachable_level of TASK but has not implemented drift_token_parameters"
        )

    @property
    def output_location_non_cached(self):
        if self.cachable_level == constants.CACHE_LEVEL_TASK:
            path = f"{escape(self.drift_token_parameters)}/{self.drift_token}"
            # path = self.drift_token
        elif self.cachable_level == constants.CACHE_LEVEL_RUN:
            path = self.run_token
        elif self.cachable_level == constants.CACHE_LEVEL_PERMANENT:
            path = "latest"
        elif self.cachable_level == constants.CACHE_LEVEL_NO_CACHE:
            path = self.run_token
        else:
            raise Exception(f"unknown cachable_level: {self.cachable_level}")

        return f"output/{self.__class__.__name__}/{self.task_reference}/{path}.json"

    @property
    def output_location_cached(self):
        return f"s3://sc-puppet-caching-bucket-{self.puppet_account_id}-{constants.HOME_REGION}/{self.output_location_non_cached}"

    @property
    def should_use_caching(self):
        return (
            self.should_use_s3_target_if_caching_is_on and config.is_caching_enabled()
        )

    @property
    def should_use_s3_target_if_caching_is_on(self):
        return self.cachable_level != constants.CACHE_LEVEL_NO_CACHE

    def output(self):
        if self.should_use_caching:
            return s3.S3Target(self.output_location_cached, format=format.UTF8)
        else:
            return luigi.LocalTarget(
                self.output_location_non_cached, format=luigi.format.Nop
            )

    def params_for_results_display(self):
        return {}

    def write_empty_output(self):
        if self.should_use_caching:
            with self.output().open("w") as f:
                f.write("{}")
        else:
            with self.output().open("wb") as f:
                f.write(b"{}")

    def write_output(self, content, skip_json_encode=False):
        if self.should_use_caching:
            with self.output().open("w") as f:
                if skip_json_encode:
                    f.write(content.decode("utf-8"))
                else:
                    f.write(serialisation_utils.json_dumps(content).decode("utf-8"))
        else:
            with self.output().open("wb") as f:
                if skip_json_encode:
                    f.write(content)
                else:
                    f.write(serialisation_utils.json_dumps(content))



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import configparser
import functools
import json
import logging
import os
import re
from copy import deepcopy

import click
import networkx as nx
from betterboto import client as betterboto_client
from deepmerge import always_merger

from servicecatalog_puppet import config, constants, serialisation_utils
from servicecatalog_puppet.macros import macros
from servicecatalog_puppet.serialisation_utils import unwrap


logger = logging.getLogger(__file__)


def get_intrinsic_functions_map(manifest_name, puppet_account_id):
    replacements = dict(AWS=dict(PuppetAccountId=puppet_account_id))
    for intrinsic_function_file in [
        manifest_name.replace("manifest.yaml", "intrinsic-functions.properties"),
        manifest_name.replace(
            "manifest.yaml", f"intrinsic-functions-{puppet_account_id}.properties"
        ),
    ]:
        parser = configparser.SafeConfigParser(
            interpolation=configparser.BasicInterpolation()
        )
        parser.optionxform = str
        if os.path.exists(intrinsic_function_file):
            logger.info(f"reading {intrinsic_function_file}")
            intrinsic_function_file_content = (
                "[Custom]\n" + open(intrinsic_function_file).read()
            )
            parser.read_string(intrinsic_function_file_content)
            for section_name, section_values in parser.items():
                if section_name == "DEFAULT":
                    continue

                if replacements.get(section_name) is None:
                    replacements[section_name] = dict()

                for item_name, item_value in section_values.items():
                    replacements[section_name][item_name] = item_value
    return replacements


def interpolate_intrinsic_functions(contents, intrinsic_functions_map):
    new_contents = contents.replace("", "")
    for section_name, section in intrinsic_functions_map.items():
        for name, value in section.items():
            function_name = "${" + section_name + "::" + name + "}"
            new_contents = new_contents.replace(function_name, value)
    return new_contents


def load(f, puppet_account_id):
    manifest_name = f.name
    manifest = {
        "schema": "puppet-2019-04-01",
        "parameters": {},
        "accounts": [],
        constants.LAUNCHES: {},
        constants.STACKS: {},
        constants.SPOKE_LOCAL_PORTFOLIOS: {},
        constants.IMPORTED_PORTFOLIOS: {},
        constants.ASSERTIONS: {},
        constants.CODE_BUILD_RUNS: {},
        constants.LAMBDA_INVOCATIONS: {},
        constants.APPS: {},
        constants.WORKSPACES: {},
        constants.CFCT: {},
        constants.SERVICE_CONTROL_POLICIES: {},
        constants.SIMULATE_POLICIES: {},
        constants.TAG_POLICIES: {},
        constants.C7N_AWS_LAMBDAS: {},
    }
    intrinsic_functions_map = get_intrinsic_functions_map(
        manifest_name, puppet_account_id
    )

    contents = f.read()
    contents = interpolate_intrinsic_functions(contents, intrinsic_functions_map)
    manifest.update(serialisation_utils.load(contents))
    d = os.path.dirname(os.path.abspath(f.name))

    extendable = constants.ALL_SECTION_NAMES + ["parameters"]
    for t in extendable:
        t_path = f"{d}{os.path.sep}{t}"
        if os.path.exists(t_path):
            for f in os.listdir(t_path):
                source = f"{t_path}{os.path.sep}{f}"
                with open(source, "r") as file:
                    contents = file.read()
                    contents = interpolate_intrinsic_functions(
                        contents, intrinsic_functions_map
                    )
                    new = serialisation_utils.load(contents)
                    for n, v in new.items():
                        if manifest[t].get(n):
                            raise Exception(f"{source} declares a duplicate {t}: {n}")
                    manifest[t].update(new)

    if os.path.exists(f"{d}{os.path.sep}manifests"):
        for f in os.listdir(f"{d}{os.path.sep}manifests"):
            with open(f"{d}{os.path.sep}manifests{os.path.sep}{f}", "r") as file:
                contents = file.read()
                contents = interpolate_intrinsic_functions(
                    contents, intrinsic_functions_map
                )
                ext = serialisation_utils.load(contents)
                for t in extendable:
                    manifest[t].update(ext.get(t, {}))

    if os.path.exists(f"{d}{os.path.sep}capabilities"):
        for f in os.listdir(f"{d}{os.path.sep}capabilities"):
            with open(f"{d}{os.path.sep}capabilities{os.path.sep}{f}", "r") as file:
                contents = file.read()
                contents = interpolate_intrinsic_functions(
                    contents, intrinsic_functions_map
                )
                ext = serialisation_utils.load(contents)
                always_merger.merge(manifest, ext)

    for config_file in [
        manifest_name.replace(".yaml", ".properties"),
        manifest_name.replace(".yaml", f"-{puppet_account_id}.properties"),
    ]:
        parser = configparser.ConfigParser(
            interpolation=configparser.BasicInterpolation()
        )
        parser.optionxform = str

        if os.path.exists(config_file):
            logger.info(f"reading {config_file}")
            parser.read(config_file)

            for section_name, section_values in parser.items():
                if section_name == "DEFAULT":
                    continue
                for item_name, item_value in section_values.items():
                    name, property_name = item_name.split(".")
                    if property_name not in ["version", "condition", "status"]:
                        raise Exception(
                            "You can only specify a version in the properties file"
                        )
                    if manifest.get(section_name, {}).get(name):
                        manifest[section_name][name][property_name] = item_value
                    else:
                        logger.warning(
                            f"Could not find manifest[{section_name}][{name}]"
                        )

    for section in constants.ALL_SPOKE_EXECUTABLE_SECTION_NAMES:
        for name, details in manifest.get(section, {}).items():
            if details.get("execution") is None:
                details["execution"] = constants.EXECUTION_MODE_DEFAULT
    return manifest


@functools.lru_cache(maxsize=32)
def get_ssm_external_account_overrides():
    result = dict()
    with betterboto_client.ClientContextManager("ssm") as ssm:
        paginator = ssm.get_paginator("get_parameters_by_path")
        for page in paginator.paginate(
            Path="/servicecatalog-puppet/manifest-external-account-overrides",
            Recursive=True,
        ):
            for parameter in page.get("Parameters", []):
                _, _, _, account_id, attribute = parameter.get("Name").split("/")
                if not result.get(account_id):
                    result[account_id] = dict()
                result[account_id][attribute] = parameter.get("Value")

    return result


def get_ssm_external_account_overrides_for(account_id, value, fail_if_missing):
    result = get_ssm_external_account_overrides().get(account_id, {})
    if fail_if_missing and not result:
        raise Exception(
            f"There is no override for {account_id} - path /servicecatalog-puppet/manifest-external-account-overrides/{account_id}/xxxx is missing"
        )
    result = result.get(value)
    if fail_if_missing and not result:
        raise Exception(
            f"There is no override for {value}.{account_id}. SSM parameter /servicecatalog-puppet/manifest-external-account-overrides/{account_id}/{value} is missing"
        )
    return result


def expand_manifest(manifest, orgs_client):
    new_manifest = deepcopy(manifest)
    temp_accounts = []
    conversions = dict()

    logger.info("Starting the expand")

    for account in manifest.get("accounts"):
        if account.get("account_id"):
            account_id = account.get("account_id")
            logger.info("Found an account: {}".format(account_id))
            expanded_account = expand_account(
                account, orgs_client, account_id, manifest
            )
            if expanded_account is None:
                raise Exception(
                    f"You have listed account: {account_id} which is not ACTIVE"
                )
            temp_accounts.append(expanded_account)
        elif account.get("ou"):
            ou = account.get("ou")
            logger.info("Found an ou: {}".format(ou))
            if ou.startswith("/"):
                ou = orgs_client.convert_path_to_ou(account.get("ou"))
                conversions[account.get("ou")] = ou
                account["ou_name"] = account["ou"]
                account["ou"] = ou
            temp_accounts += expand_ou(account, orgs_client, manifest)

    for parameter_name, parameter_details in new_manifest.get("parameters", {}).items():
        if parameter_details.get("macro"):
            macro_to_run = macros.get(parameter_details.get("macro").get("method"))
            result = macro_to_run(
                orgs_client, parameter_details.get("macro").get("args")
            )
            parameter_details["default"] = result
            del parameter_details["macro"]

    accounts_by_id = {}
    for account in temp_accounts:
        for parameter_name, parameter_details in account.get("parameters", {}).items():
            if parameter_details.get("macro"):
                macro_to_run = macros.get(parameter_details.get("macro").get("method"))
                result = macro_to_run(
                    orgs_client, parameter_details.get("macro").get("args")
                )
                parameter_details["default"] = result
                del parameter_details["macro"]

        account_id = account.get("account_id")
        if account.get("append") or account.get("overwrite"):
            if (
                account.get("default_region")
                or account.get("regions_enabled")
                or account.get("tags")
            ):
                raise Exception(
                    f"{account_id}: If using append or overwrite you cannot set default_region, regions_enabled or tags"
                )

        if accounts_by_id.get(account_id) is None:
            accounts_by_id[account_id] = account
        else:
            logger.debug("Found an already seen account")
            if not (account.get("append") or account.get("overwrite")):
                raise Exception(
                    f"Account {account_id} was seen more than once without an append or overwrite"
                )
            accounts_by_id[account_id].update(account)

    for stored_account_id, stored_account in accounts_by_id.items():
        # Get tags from orgs if we should
        organizations_account_tags = stored_account.get(
            "organizations_account_tags", "ignored"
        )
        if organizations_account_tags != "ignored":
            tags = list()
            paginator = orgs_client.get_paginator("list_tags_for_resource")
            for page in paginator.paginate(
                ResourceId=stored_account_id,
            ):
                tags.extend(page.get("Tags", []))
            tags = [f"{t['Key']}:{t['Value']}" for t in tags]

            if organizations_account_tags == "append":
                stored_account["tags"] = stored_account.get("tags", []) + tags
            elif organizations_account_tags == "honour":
                stored_account["tags"] = tags

        # Get tags from orgs if we should
        external_account_overrides = stored_account.get(
            "external_account_overrides", {}
        )
        if external_account_overrides.get("default_region"):
            ####
            if external_account_overrides.get("default_region").get("source") == "ssm":
                default_region = get_ssm_external_account_overrides_for(
                    stored_account.get("account_id"),
                    "default_region",
                    external_account_overrides.get("default_region").get(
                        "fail_if_missing", True
                    ),
                )
                if default_region:
                    stored_account["default_region"] = default_region
            if external_account_overrides.get("enabled_regions").get("source") == "ssm":
                enabled_regions = get_ssm_external_account_overrides_for(
                    stored_account.get("account_id"),
                    "enabled_regions",
                    external_account_overrides.get("enabled_regions").get(
                        "fail_if_missing", True
                    ),
                )
                if enabled_regions:
                    stored_account["enabled_regions"] = serialisation_utils.load(
                        enabled_regions
                    )

        # append or overwrite if we should
        if stored_account.get("overwrite"):
            overwrite = stored_account.get("overwrite")
            if overwrite.get("tags"):
                stored_account["tags"] = overwrite.get("tags")
            if overwrite.get("regions_enabled"):
                stored_account["regions_enabled"] = overwrite.get("regions_enabled")
            if overwrite.get("default_region"):
                stored_account["default_region"] = overwrite.get("default_region")
            del stored_account["overwrite"]

        elif stored_account.get("append"):
            append = stored_account.get("append")
            for tag in append.get("tags", []):
                stored_account["tags"] = stored_account.get("tags", []) + [tag]
            for region_enabled in append.get("regions_enabled", []):
                stored_account["regions_enabled"] = stored_account.get(
                    "regions_enabled", []
                ) + [region_enabled]
            del stored_account["append"]

    new_manifest["accounts"] = list(accounts_by_id.values())

    for section in [constants.LAUNCHES, constants.STACKS]:
        for name, details in new_manifest.get(section, {}).items():
            for parameter_name, parameter_details in details.get(
                "parameters", {}
            ).items():
                if parameter_details.get("macro"):
                    macro_to_run = macros.get(
                        parameter_details.get("macro").get("method")
                    )
                    result = macro_to_run(
                        orgs_client, parameter_details.get("macro").get("args")
                    )
                    parameter_details["default"] = result
                    del parameter_details["macro"]

    return conversions, new_manifest


def rewrite_deploy_as_share_to_for_spoke_local_portfolios(manifest):
    for item_name, item in manifest.get(constants.SPOKE_LOCAL_PORTFOLIOS, {}).items():
        if item.get("deploy_to"):
            item["share_with"] = item["deploy_to"]
            del item["deploy_to"]
    return manifest


def rewrite_cfct(manifest):
    manifest_accounts = dict()
    for account in manifest.get("accounts", []):
        if account.get("account_id"):
            manifest_accounts[account.get("name")] = account.get("account_id")

    prev = None
    for instance in manifest.get(constants.CFCT, []):
        if str(instance.get("version")) != "2021-03-15":
            raise Exception(
                f"not supported version of cfct manifest {instance.get('version')}"
            )
        default_region = instance.get("region")
        for resource in instance.get("resources", []):
            resource_file = resource.get("resource_file").lower()
            if resource_file.startswith("s3://"):
                m = re.match(r"s3://(.*)/(.*)", resource_file)
                bucket = m.group(1)
                key = m.group(2)
            elif resource_file.startswith("https://"):
                m = re.match(r"https://([a-z0-9-]+)(.*)/(.*)", resource_file)
                bucket = m.group(1)
                key = m.group(3)
            else:
                raise Exception(
                    f"All resource files should begin with s3:// of https://: {resource_file}"
                )

            name = resource.get("name")
            deploy_method = resource.get("deploy_method")

            if deploy_method == "stack_set":
                parameters = dict()
                depends_on = list()
                ssm = list()
                outputs = dict(ssm=ssm)
                deploy_to_accounts = list()
                deploy_to_tags = list()
                deploy_to = dict(tags=deploy_to_tags, accounts=deploy_to_accounts)

                if resource.get("parameter_file"):
                    parameter_file = resource.get("parameter_file")
                    if parameter_file.startswith("s3://"):
                        m = re.match(r"s3://(.*)/(.*)", parameter_file)
                        bucket = m.group(1)
                        key = m.group(2)
                    elif parameter_file.startswith("https://"):
                        m = re.match(r"https://([a-z0-9-]+)(.*)/(.*)", parameter_file)
                        bucket = m.group(1)
                        key = m.group(3)
                    else:
                        raise Exception(
                            f"All parameter_files should begin with s3:// of https://: {parameter_file}"
                        )
                    with betterboto_client.ClientContextManager("s3") as s3:
                        p = s3.get_object(Bucket=bucket, Key=key).read()
                        resource["parameters"] = serialisation_utils.json_loads(p)

                for p in resource.get("parameters", []):
                    parameter_key = p.get("parameter_key")
                    parameter_value = p.get("parameter_value")
                    m = re.match(r"\$\[alfred_ssm_(.*)\]", parameter_value)
                    if m:
                        parameters[parameter_key] = dict(
                            ssm=dict(
                                name=m.group(1),
                                region=default_region,
                            )
                        )
                    else:
                        parameters[parameter_key] = dict(default=parameter_value)

                if prev is not None:
                    depends_on.append(
                        dict(
                            name=prev,
                            type=constants.STACK,
                            affinity=constants.STACK,
                        )
                    )

                for output in resource.get("export_outputs", []):
                    output_value = re.match(
                        r"\$\[output_(.*)\]", output.get("value")
                    ).group(1)
                    ssm.append(
                        dict(param_name=output.get("name"), stack_output=output_value)
                    )

                regions = resource.get("regions", [default_region])
                for account in resource.get("deployment_targets", {}).get(
                    "accounts", []
                ):
                    if re.match(r"[0-9]{12}", str(account)):
                        deploy_to_accounts.append(
                            dict(account_id=account, regions=regions)
                        )
                    else:
                        if manifest_accounts.get(account) is None:
                            raise Exception(
                                f"You are using CFCT resource: {name} to deploy to account: {account} which is not defined in your accounts section"
                            )
                        deploy_to_accounts.append(
                            dict(
                                account_id=manifest_accounts.get(account),
                                regions=regions,
                            )
                        )

                for organizational_unit in resource.get("deployment_targets", {}).get(
                    "organizational_units", []
                ):
                    deploy_to_tags.append(
                        dict(
                            tag=f"autogenerated:{organizational_unit}", regions=regions
                        )
                    )

                stack = dict(
                    name=name,
                    stack_set_name=name,
                    bucket=bucket,
                    key=key,
                    execution=constants.EXECUTION_MODE_HUB,
                    capabilities=["CAPABILITY_NAMED_IAM"],
                    parameters=parameters,
                    depends_on=depends_on,
                    outputs=outputs,
                    deploy_to=deploy_to,
                )

                if manifest.get(constants.STACKS) is None:
                    manifest[constants.STACKS] = dict()
                if manifest[constants.STACKS].get(name) is not None:
                    raise Exception(
                        f"You have a stack and a cfct resource with the same name: {name}"
                    )
                manifest[constants.STACKS][name] = stack

                prev = name

            elif deploy_method == "scp":
                with betterboto_client.ClientContextManager("s3") as s3:
                    p = s3.get_object(Bucket=bucket, Key=key).read()
                    content = serialisation_utils.json_loads(p)

                depends_on = list()
                deploy_to_accounts = list()
                deploy_to_ous = list()
                deploy_to = dict(accounts=deploy_to_accounts, ous=deploy_to_ous)

                if prev is not None:
                    depends_on.append(
                        dict(
                            name=prev,
                            type=constants.SERVICE_CONTROL_POLICY,
                            affinity=constants.SERVICE_CONTROL_POLICY,
                        )
                    )

                regions = "home_region"
                for account in resource.get("deployment_targets", {}).get(
                    "accounts", []
                ):
                    if re.match(r"[0-9]{12}", str(account)):
                        deploy_to_accounts.append(
                            dict(account_id=account, regions=regions)
                        )
                    else:
                        if manifest_accounts.get(account) is None:
                            raise Exception(
                                f"You are using CFCT resource: {name} to deploy to account: {account} which is not defined in your accounts section"
                            )
                        deploy_to_accounts.append(
                            dict(
                                account_id=manifest_accounts.get(account),
                                regions=regions,
                            )
                        )

                for organizational_unit in resource.get("deployment_targets", {}).get(
                    "organizational_units", []
                ):
                    deploy_to_ous.append(dict(ou=organizational_unit, regions=regions))

                scp = dict(
                    description=resource.get("description", "auto generated from CfCT"),
                    content=dict(default=content),
                    depends_on=depends_on,
                    apply_to=deploy_to,
                )

                if manifest.get(constants.SERVICE_CONTROL_POLICIES) is None:
                    manifest[constants.SERVICE_CONTROL_POLICIES] = dict()
                if manifest[constants.SERVICE_CONTROL_POLICIES].get(name) is not None:
                    raise Exception(
                        f"You have an SCP and a cfct resource with the same name: {name}"
                    )
                manifest[constants.SERVICE_CONTROL_POLICIES][name] = scp

                prev = name

            else:
                raise Exception(f"Unknown deploy_method of {deploy_method}")

    return manifest


def rewrite_depends_on(manifest):
    for (
        section_item_name,
        section_name,
    ) in constants.ALL_SECTION_NAME_SINGULAR_AND_PLURAL_LIST:
        for item, details in manifest.get(section_name, {}).items():
            for i in range(len(details.get("depends_on", []))):
                if isinstance(details["depends_on"][i], str):
                    manifest[section_name][item]["depends_on"][i] = dict(
                        name=details["depends_on"][i],
                        type="launch",
                    )
                if isinstance(details["depends_on"][i], dict):
                    if details["depends_on"][i].get(constants.AFFINITY) is None:
                        details["depends_on"][i][constants.AFFINITY] = details[
                            "depends_on"
                        ][i]["type"]
    return manifest


def rewrite_ssm_parameters(manifest):
    """
    when an item in a section of the manifest uses an ssm parameter this will add a depends on to the ssm parameter
    where it finds the parameter being set up by the output of a dependency.
    :param manifest:
    :return:
    """
    for (
        section_item_name,
        section_name,
    ) in constants.SECTION_NAME_SINGULAR_AND_PLURAL_LIST_THAT_SUPPORTS_PARAMETERS:
        for item, details in manifest.get(section_name, {}).items():
            for parameter_name, parameter_details in details.get(
                "parameters", {}
            ).items():
                if parameter_details.get("cloudformation_stack_output"):
                    existing_parameter = parameter_details.get(
                        "cloudformation_stack_output"
                    )
                    stack_name = existing_parameter.get("stack_name")
                    output_key = existing_parameter.get("output_key")
                    new_parameter = dict(
                        account_id=existing_parameter.get(
                            "account_id", "${AWS::AccountId}"
                        ),
                        region=existing_parameter.get(
                            "region", "${AWS::PuppetAccountId}"
                        ),
                        client="cloudformation",
                        call="describe_stacks",
                        use_paginator=True,
                        arguments=dict(StackName=stack_name),
                        filter=f"Stacks[?StackStatus==`CREATE_COMPLETE` || StackStatus==`UPDATE_COMPLETE`|| StackStatus==`UPDATE_ROLLBACK_COMPLETE`].Outputs[] | [?OutputKey==`{output_key}`].OutputValue | [0]",
                    )
                    parameter_details["boto3"] = new_parameter

                if parameter_details.get("servicecatalog_provisioned_product_output"):
                    existing_parameter = parameter_details.get(
                        "servicecatalog_provisioned_product_output"
                    )
                    provisioned_product_name = existing_parameter.get(
                        "provisioned_product_name"
                    )
                    output_key = existing_parameter.get("output_key")
                    new_parameter = dict(
                        account_id=existing_parameter.get(
                            "account_id", "${AWS::AccountId}"
                        ),
                        region=existing_parameter.get("region", "${AWS::Region}"),
                        client="servicecatalog",
                        call="get_provisioned_product_outputs",
                        use_paginator=False,
                        arguments=dict(ProvisionedProductName=provisioned_product_name),
                        filter=f"Outputs[?OutputKey==`{output_key}`].OutputValue | [0]",
                    )
                    parameter_details["boto3"] = new_parameter
    return manifest


def rewrite_stacks(manifest, puppet_account_id):
    for category, section in [
        (constants.STACK, constants.STACKS),
        (constants.APP, constants.APPS),
        (constants.WORKSPACE, constants.WORKSPACES),
    ]:
        for item, details in manifest.get(section, {}).items():
            if not details.get("key"):
                if category == constants.STACK:
                    details["key"] = (
                        f"{category}/{details['name']}/{details['version']}/{category}.template-${{AWS::Region}}.yaml"
                    )
                else:
                    details["key"] = (
                        f"{category}/{details['name']}/{details['version']}/{category}.zip"
                    )
                del details["name"]
                del details["version"]
                if category == constants.STACK:
                    if (
                        details.get(constants.MANIFEST_SHOULD_USE_STACKS_SERVICE_ROLE)
                        is None
                    ):
                        details[constants.MANIFEST_SHOULD_USE_STACKS_SERVICE_ROLE] = (
                            config.get_should_use_stacks_service_role(puppet_account_id)
                        )
    return manifest


def rewrite_scps(manifest, puppet_account_id):
    for item, details in manifest.get(constants.SERVICE_CONTROL_POLICIES, {}).items():
        for attribute in ["tags", "accounts", "ous"]:
            apply_to = details.get("apply_to")
            for d in apply_to.get(attribute, []):
                d["regions"] = "home_region"
    for item, details in manifest.get(constants.TAG_POLICIES, {}).items():
        for attribute in ["tags", "accounts", "ous"]:
            apply_to = details.get("apply_to")
            for d in apply_to.get(attribute, []):
                d["regions"] = "home_region"
    return manifest


def expand_path(account, client, manifest):
    ou = client.convert_path_to_ou(account.get("ou"))
    account["ou_name"] = account["ou"]
    account["ou"] = ou
    return expand_ou(account, client, manifest)


def expand_account(account, client, account_id, manifest):
    response = client.describe_account(AccountId=account_id)

    new_account = deepcopy(manifest.get("defaults", {}).get("accounts", {}))
    new_account.update(deepcopy(account))

    ou_from_parent = None
    if "ou" in new_account:
        ou_from_parent = new_account["ou"]
        del new_account["ou"]

    account_details = response.get("Account")
    if account_details.get("Status") == "ACTIVE":
        if account_details.get("Name") is not None:
            new_account["name"] = account_details.get("Name")
        new_account["email"] = account_details.get("Email")
        if ou_from_parent is not None:
            new_account["expanded_from"] = ou_from_parent
            new_account["account_id"] = account_id
        new_account["organization"] = (
            account_details.get("Arn").split(":")[5].split("/")[1]
        )
        return new_account
    else:
        logger.info(
            f"Skipping account as it is not ACTIVE: {json.dumps(account_details, default=str)}"
        )
    return None


def expand_ou(original_account, client, manifest):
    expanded = []
    exclusions = original_account.get("exclude", {}).get("accounts", [])
    ou_exclusions = original_account.get("exclude", {}).get("ous", [])
    for ou_exclusion in ou_exclusions:
        if ou_exclusion.startswith("/"):
            ou_id = client.convert_path_to_ou(ou_exclusion)
        else:
            ou_id = ou_exclusion
        children = client.list_children_nested(ParentId=ou_id, ChildType="ACCOUNT")
        for child in children:
            logger.info(
                f"Adding {child.get('Id')} to the exclusion list as it was in the ou {ou_exclusion}"
            )
            exclusions.append(child.get("Id"))

    response = client.list_children_nested(
        ParentId=original_account.get("ou"), ChildType="ACCOUNT"
    )
    for result in response:
        new_account_id = result.get("Id")
        if new_account_id in exclusions:
            logger.info(f"Skipping {new_account_id} as it is in the exclusion list")
            continue
        new_account = expand_account(original_account, client, new_account_id, manifest)
        if new_account:
            expanded.append(new_account)
    return expanded


def get_configuration_overrides(manifest, launch_details):
    configuration = dict()
    if manifest.get("configuration"):
        if manifest.get("configuration").get("retry_count"):
            configuration["retry_count"] = manifest.get("configuration").get(
                "retry_count"
            )
    if launch_details.get("configuration"):
        if launch_details.get("configuration").get("retry_count"):
            configuration["retry_count"] = launch_details.get("configuration").get(
                "retry_count"
            )
        if launch_details.get("configuration").get("requested_priority"):
            configuration["requested_priority"] = int(
                launch_details.get("configuration").get("requested_priority")
            )
    return configuration


def get_from_dict(d, path):
    parts = path.split("/")
    if len(parts) == 1:
        result = d.get(parts[0])
    elif len(parts) == 2:
        result = d.get(parts[0], {}).get(parts[1])
    elif len(parts) == 3:
        result = d.get(parts[0], {}).get(parts[1], {}).get(parts[2])
    if result is None:
        raise KeyError
    return result


class Manifest(dict):
    def has_cache(self):
        return self.get("id_cache") is not None

    def get_launches_items(self):
        return self.get(constants.LAUNCHES, {}).items()

    def get_launch(self, name):
        return self.get(constants.LAUNCHES).get(name)

    def get_workspace(self, name):
        return self.get(constants.WORKSPACES).get(name)

    def get_app(self, name):
        return self.get(constants.APPS).get(name)

    def get_tasks_for(
        self,
        puppet_account_id,
        section_name,
        item_name,
        home_region_configured,
        regions_configured,
        single_account="None",
    ):
        accounts = self.get(constants.ACCOUNTS)
        section = self.get(section_name)
        provisioning_tasks = list()
        item = section[item_name]
        sharing_mode_default = config.get_global_sharing_mode_default()
        share_tag_options_default = config.get_global_share_tag_options_default()
        share_principals_default = config.get_global_share_principals_default()

        deploy_to = {
            "launches": "deploy_to",
            "stacks": "deploy_to",
            "apps": "deploy_to",
            "workspaces": "deploy_to",
            "spoke-local-portfolios": "share_with",
            "imported-portfolios": "share_with",
            "lambda-invocations": "invoke_for",
            "code-build-runs": "run_for",
            "assertions": "assert_for",
            "service-control-policies": "apply_to",
            "tag-policies": "apply_to",
            "simulate-policies": "simulate_for",
            constants.ORGANIZATIONAL_UNITS: "create_in",
            constants.C7N_AWS_LAMBDAS: "apply_to",
        }.get(section_name)

        if (
            section_name == constants.SPOKE_LOCAL_PORTFOLIOS
            and item.get(deploy_to) is None
        ):
            deploy_to = "deploy_to"

        common_parameters = {
            "launches": dict(
                launch_name=item_name,
                # launch_parameters=item.get("parameters", {}),
                # manifest_parameters=self.get("parameters", {}),
                ssm_param_outputs=item.get("outputs", {}).get("ssm", []),
                share_tag_options=item.get(
                    "share_tag_options", share_tag_options_default
                ),
                share_principals=item.get("share_principals", share_principals_default),
                portfolio=item.get("portfolio"),
                product=item.get("product"),
                version=item.get("version"),
                execution=item.get("execution", constants.EXECUTION_MODE_DEFAULT),
                tags=item.get("tags", []),
            ),
            "stacks": dict(
                stack_name=item_name,
                launch_name=item.get("launch_name", ""),
                stack_set_name=item.get("stack_set_name", ""),
                # launch_parameters=item.get("parameters", {}),
                capabilities=item.get("capabilities", []),
                # manifest_parameters=self.get("parameters", {}),
                ssm_param_outputs=item.get("outputs", {}).get("ssm", []),
                bucket=f"sc-puppet-stacks-repository-{puppet_account_id}",
                key=item.get("key"),
                version_id=item.get("version_id", ""),
                execution=item.get("execution", constants.EXECUTION_MODE_DEFAULT),
                use_service_role=item.get(
                    constants.MANIFEST_SHOULD_USE_STACKS_SERVICE_ROLE,
                    constants.CONFIG_SHOULD_USE_STACKS_SERVICE_ROLE_DEFAULT,
                ),
                tags=item.get("tags", []),
            ),
            "apps": dict(
                app_name=item_name,
                # launch_parameters=item.get("parameters", {}),
                # manifest_parameters=self.get("parameters", {}),
                ssm_param_outputs=item.get("outputs", {}).get("ssm", []),
                bucket=f"sc-puppet-stacks-repository-{puppet_account_id}",
                key=item.get("key"),
                version_id=item.get("version_id", ""),
                execution=item.get("execution", constants.EXECUTION_MODE_DEFAULT),
            ),
            "workspaces": dict(
                workspace_name=item_name,
                # launch_parameters=item.get("parameters", {}),
                # manifest_parameters=self.get("parameters", {}),
                ssm_param_outputs=item.get("outputs", {}).get("ssm", []),
                bucket=f"sc-puppet-stacks-repository-{puppet_account_id}",
                key=item.get("key"),
                version_id=item.get("version_id", ""),
                execution=item.get("execution", constants.EXECUTION_MODE_DEFAULT),
            ),
            "spoke-local-portfolios": dict(
                spoke_local_portfolio_name=item_name,
                product_generation_method=item.get(
                    "product_generation_method",
                    constants.PRODUCT_GENERATION_METHOD_DEFAULT,
                ),
                execution=item.get("execution", constants.EXECUTION_MODE_DEFAULT),
                sharing_mode=item.get("sharing_mode", sharing_mode_default),
                share_tag_options=item.get(
                    "share_tag_options", share_tag_options_default
                ),
                share_principals=item.get("share_principals", share_principals_default),
                associations=item.get("associations", list()),
                launch_constraints=item.get("constraints", {}).get("launch", []),
                resource_update_constraints=item.get("constraints", {}).get(
                    "resource_update", []
                ),
                portfolio=item.get("portfolio"),
            ),
            "imported-portfolios": dict(
                imported_portfolio_name=item_name,
                execution=item.get("execution", constants.EXECUTION_MODE_DEFAULT),
                sharing_mode=item.get("sharing_mode", sharing_mode_default),
                was_a_spoke_local_portfolio=item.get(
                    "was_a_spoke_local_portfolio",
                    constants.WAS_A_SPOKE_LOCAL_PORTFOLIO_DEFAULT,
                ),
                share_tag_options=item.get(
                    "share_tag_options", share_tag_options_default
                ),
                share_principals=item.get("share_principals", share_principals_default),
                associations=item.get("associations", list()),
                portfolio=item.get("portfolio"),
            ),
            "lambda-invocations": dict(
                lambda_invocation_name=item_name,
                function_name=item.get("function_name"),
                qualifier=item.get("qualifier", "$LATEST"),
                invocation_type=item.get("invocation_type"),
                # launch_parameters=item.get("parameters", {}),
                # manifest_parameters=self.get("parameters", {}),
                execution=item.get("execution", constants.EXECUTION_MODE_DEFAULT),
            ),
            "code-build-runs": dict(
                code_build_run_name=item_name,
                # launch_parameters=item.get("parameters", {}),
                # manifest_parameters=self.get("parameters", {}),
                project_name=item.get("project_name"),
                execution=item.get("execution", constants.EXECUTION_MODE_DEFAULT),
            ),
            "assertions": dict(
                assertion_name=item_name,
                expected=item.get("expected"),
                actual=item.get("actual"),
                execution=item.get("execution", constants.EXECUTION_MODE_DEFAULT),
            ),
            "service-control-policies": dict(
                service_control_policy_name=item_name,
                description=item.get("description"),
                content=unwrap(item.get("content")),
            ),
            "tag-policies": dict(
                tag_policy_name=item_name,
                description=item.get("description"),
                content=unwrap(item.get("content")),
            ),
            constants.SIMULATE_POLICIES: dict(
                execution=item.get("execution", constants.EXECUTION_MODE_DEFAULT),
                simulate_policy_name=item_name,
                simulation_type=item.get("simulation_type"),
                policy_source_arn=item.get("policy_source_arn", ""),
                policy_input_list=item.get("policy_input_list", []),
                permissions_boundary_policy_input_list=item.get(
                    "permissions_boundary_policy_input_list", []
                ),
                action_names=item.get("action_names"),
                expected_decision=item.get("expected_decision"),
                resource_arns=item.get("resource_arns", []),
                resource_policy=item.get("resource_policy", ""),
                resource_owner=item.get("resource_owner", ""),
                caller_arn=item.get("caller_arn", ""),
                context_entries=item.get("context_entries", []),
                resource_handling_option=item.get("resource_handling_option", ""),
            ),
            constants.ORGANIZATIONAL_UNITS: dict(
                organizational_unit_name=item_name,
                path=item.get("path"),
                parent_ou_id=item.get("parent_ou_id"),
                name=item.get("name"),
                tags=item.get("tags"),
            ),
            constants.C7N_AWS_LAMBDAS: dict(
                execution=item.get("execution", constants.EXECUTION_MODE_DEFAULT),
                policies=item.get("policies"),
                custodian=item.get("custodian"),
                role_name=item.get(
                    "role_name", constants.C7N_CUSTODIAN_ROLE_NAME_DEFAULT
                ),
                role_path=item.get(
                    "role_path", constants.C7N_CUSTODIAN_ROLE_PATH_DEFAULT
                ),
                schedule_expression=item.get("schedule_expression", ""),
                role_managed_policy_arns=item.get(
                    "role_managed_policy_arns",
                    constants.C7N_CUSTODIAN_MANAGED_POLICY_ARNS_DEFAULT,
                ),
                c7n_version=item.get("c7n_version", constants.C7N_VERSION_DEFAULT),
                c7n_org_version=item.get("c7n_org_version", ""),
            ),
        }.get(section_name)

        common_parameters.update(
            dict(
                puppet_account_id=puppet_account_id,
                status=item.get("status"),
                requested_priority=item.get("requested_priority", 0),
                dependencies=item.get("depends_on", []),
            )
        )

        # handle deploy_to tags
        tags = item.get(deploy_to).get("tags", [])
        for tag in tags:
            tag_name = tag.get("tag")
            regions = tag.get("regions")
            for account in accounts:
                account_id = str(account.get("account_id"))
                if single_account != "None" and single_account != account_id:
                    continue

                if tag_name in account.get("tags"):
                    additional_parameters = {
                        "launches": dict(
                            account_id=account_id,
                            ou=account.get("expanded_from", ""),
                        ),
                        "apps": dict(
                            account_id=account_id,
                        ),
                        "workspaces": dict(
                            account_id=account_id,
                        ),
                        "stacks": dict(
                            account_id=account_id,
                        ),
                        "spoke-local-portfolios": dict(
                            account_id=account_id,
                            organization=account.get("organization", ""),
                            ou=account.get("expanded_from", ""),
                        ),
                        "imported-portfolios": dict(
                            account_id=account_id,
                            organization=account.get("organization", ""),
                            ou=account.get("expanded_from", ""),
                        ),
                        "assertions": dict(
                            account_id=account_id,
                        ),
                        "lambda-invocations": dict(
                            account_id=account_id,
                        ),
                        "code-build-runs": dict(
                            account_id=account_id,
                        ),
                        "service-control-policies": dict(
                            account_id=account_id,
                            ou_name="",
                        ),
                        "tag-policies": dict(
                            account_id=account_id,
                            ou_name="",
                        ),
                        constants.SIMULATE_POLICIES: dict(
                            account_id=account_id,
                        ),
                        constants.C7N_AWS_LAMBDAS: dict(
                            account_id=account_id,
                        ),
                    }.get(section_name)
                    if isinstance(regions, str):
                        if regions in [
                            "enabled",
                            "regions_enabled",
                            "enabled_regions",
                        ]:
                            for region_enabled in account.get("regions_enabled"):
                                provisioning_tasks.append(
                                    dict(
                                        **common_parameters,
                                        **additional_parameters,
                                        region=region_enabled,
                                    )
                                )
                        elif regions == "default_region":
                            provisioning_tasks.append(
                                dict(
                                    **common_parameters,
                                    **additional_parameters,
                                    region=account.get("default_region"),
                                )
                            )
                        elif regions == "home_region":
                            provisioning_tasks.append(
                                dict(
                                    **common_parameters,
                                    **additional_parameters,
                                    region=home_region_configured,
                                )
                            )
                        elif regions == "all":
                            all_regions = regions_configured
                            for region_enabled in all_regions:
                                provisioning_tasks.append(
                                    dict(
                                        **common_parameters,
                                        **additional_parameters,
                                        region=region_enabled,
                                    )
                                )

                        else:
                            raise Exception(
                                f"Unsupported regions {regions} setting for {section_name}: {item_name}"
                            )
                    elif isinstance(regions, list):
                        for region_ in regions:
                            provisioning_tasks.append(
                                dict(
                                    **common_parameters,
                                    **additional_parameters,
                                    region=region_,
                                )
                            )
                    elif isinstance(regions, tuple):
                        for region_ in regions:
                            provisioning_tasks.append(
                                dict(
                                    **common_parameters,
                                    **additional_parameters,
                                    region=region_,
                                )
                            )

                    else:
                        raise Exception(
                            f"Unsupported regions {regions} setting for {section_name}: {item_name}"
                        )

        # handle deploy_to accounts
        for account_to_deploy_to in item.get(deploy_to).get("accounts", []):
            account_id_of_account_to_deploy_to = account_to_deploy_to.get("account_id")
            regions = account_to_deploy_to.get("regions")

            account = self.get_account(account_id_of_account_to_deploy_to)
            account_id = account_id_of_account_to_deploy_to
            if single_account != "None" and single_account != account_id:
                continue
            additional_parameters = {
                "launches": dict(
                    account_id=account_id,
                ),
                "stacks": dict(
                    account_id=account_id,
                ),
                "spoke-local-portfolios": dict(
                    account_id=account_id,
                ),
                "imported-portfolios": dict(
                    account_id=account_id,
                ),
                "assertions": dict(
                    account_id=account_id,
                ),
                "lambda-invocations": dict(
                    account_id=account_id,
                ),
                "code-build-runs": dict(
                    account_id=account_id,
                ),
                "service-control-policies": dict(
                    account_id=account_id,
                    ou_name="",
                ),
                "tag-policies": dict(
                    account_id=account_id,
                    ou_name="",
                ),
                constants.SIMULATE_POLICIES: dict(
                    account_id=account_id,
                ),
                constants.ORGANIZATIONAL_UNITS: dict(
                    account_id=account_id,
                ),
                constants.C7N_AWS_LAMBDAS: dict(
                    account_id=account_id,
                ),
            }.get(section_name)

            if isinstance(regions, str):
                if regions in [
                    "enabled",
                    "regions_enabled",
                    "enabled_regions",
                ]:
                    for region_enabled in account.get("regions_enabled"):
                        provisioning_tasks.append(
                            dict(
                                **common_parameters,
                                **additional_parameters,
                                region=region_enabled,
                            )
                        )
                elif regions == "default_region":
                    provisioning_tasks.append(
                        dict(
                            **common_parameters,
                            **additional_parameters,
                            region=account.get("default_region"),
                        )
                    )
                elif regions == "home_region":
                    provisioning_tasks.append(
                        dict(
                            **common_parameters,
                            **additional_parameters,
                            region=home_region_configured,
                        )
                    )
                elif regions == "all":
                    all_regions = regions_configured
                    for region_enabled in all_regions:
                        provisioning_tasks.append(
                            dict(
                                **common_parameters,
                                **additional_parameters,
                                region=region_enabled,
                            )
                        )

                else:
                    raise Exception(
                        f"Unsupported regions {regions} setting for {section_name}: {item_name}"
                    )
            elif isinstance(regions, list):
                for region_ in regions:
                    provisioning_tasks.append(
                        dict(
                            **common_parameters,
                            **additional_parameters,
                            region=region_,
                        )
                    )
            elif isinstance(regions, tuple):
                for region_ in regions:
                    provisioning_tasks.append(
                        dict(
                            **common_parameters,
                            **additional_parameters,
                            region=region_,
                        )
                    )

            else:
                raise Exception(
                    f"Unsupported regions {regions} setting for {section_name}: {item_name}"
                )

        # handle deploy_to ous
        for ou_to_deploy_to in item.get(deploy_to).get("ous", []):
            ou_name = ou_to_deploy_to.get("ou")
            regions = ou_to_deploy_to.get("regions")

            if single_account != "None":
                continue

            additional_parameters = {
                "service-control-policies": dict(
                    account_id="",
                    ou_name=ou_name,
                ),
                "tag-policies": dict(
                    account_id="",
                    ou_name=ou_name,
                ),
            }.get(section_name)

            if isinstance(regions, str) and regions == "home_region":
                provisioning_tasks.append(
                    dict(
                        **common_parameters,
                        **additional_parameters,
                        region=home_region_configured,
                    )
                )
            else:
                raise Exception(
                    f"Unsupported regions {regions} setting for {section_name}: {item_name}"
                )

        return provisioning_tasks

    def get_account_ids_used_for_section_item(
        self, puppet_account_id, section_name, item_name
    ):
        return list(
            set(
                task.get("account_id")
                for task in self.get_tasks_for(
                    puppet_account_id, section_name, item_name
                )
            )
        )

    def get_account_ids_and_regions_used_for_section_item(
        self, puppet_account_id, section_name, item_name
    ):
        result = dict()
        for task in self.get_tasks_for(puppet_account_id, section_name, item_name):
            if result.get(task.get("account_id")) is None:
                result[task.get("account_id")] = list()
            result[task.get("account_id")].append(task.get("region"))
        for account_id in result.keys():
            result[account_id] = list(set(result[account_id]))
        return result

    def get_mapping(self, mapping, account_id, region):
        manifest_mappings = self.get("mappings")
        new_mapping = []
        for item in mapping:
            if item == "AWS::AccountId":
                new_mapping.append(account_id)
            elif item == "AWS::Region":
                new_mapping.append(region)
            else:
                new_mapping.append(item)
        path_to_find = "/".join(new_mapping)
        result = None
        try:
            result = get_from_dict(manifest_mappings, path_to_find)
        except KeyError:
            logger.debug(f"Could not find: {path_to_find}")
            while len(new_mapping) > 0:
                new_mapping[-1] = "default"
                path_to_find = "/".join(new_mapping)
                try:
                    logger.info(f"now looking for {path_to_find}")
                    result = get_from_dict(manifest_mappings, path_to_find)
                    break
                except KeyError:
                    logger.info(f"Could not find {path_to_find}")
                    new_mapping.pop()

        if result is None:
            raise Exception(f"Could not find: {'' + '/'.join(mapping)}")
        return result

    def get_account(self, account_id):
        for account in self.get("accounts"):
            if account.get("account_id") == str(account_id):
                return account
        raise Exception(f"Could not find account: {account_id}")

    def get_parameters_for_account(self, account_id):
        for account in self.get("accounts"):
            if account.get("account_id") == str(account_id):
                return account.get("parameters", {})
        return dict()

    def get_sharing_policies_by_region(self):
        sharing_policies_by_region = {}
        for account in self.get("accounts"):
            account_regions = list()
            if account.get("default_region") is None:
                raise Exception(
                    f"Account {account.get('account_id')} has no default_region"
                )
            account_regions.append(account.get("default_region"))

            enabled_regions = (
                account.get("enabled", [])
                + account.get("regions_enabled", [])
                + account.get("enabled_regions", [])
            )
            if len(enabled_regions) == 0:
                raise Exception(
                    f"Account {account.get('account_id')} has no enabled|regions_enabled|enabled_regions"
                )
            account_regions += enabled_regions

            for r in account_regions:
                if sharing_policies_by_region.get(r) is None:
                    sharing_policies_by_region[r] = dict(accounts=[], organizations=[])
                if account.get("organization", "") != "":
                    organization = account.get("organization")
                    if (
                        organization
                        not in sharing_policies_by_region[r]["organizations"]
                    ):
                        sharing_policies_by_region[r]["organizations"].append(
                            organization
                        )
                else:
                    account_id = account.get("account_id")
                    if account_id not in sharing_policies_by_region[r]["accounts"]:
                        sharing_policies_by_region[r]["accounts"].append(account_id)

        return sharing_policies_by_region

    def get_shares_by_region_portfolio_account(self, puppet_account_id, section):
        shares_by_region_portfolio_account = {}
        configuration = {}

        for launch_name, launch_details in self.get(section, {}).items():
            portfolio = launch_details.get("portfolio")
            tasks = self.get_task_defs_from_details(
                puppet_account_id,
                launch_name,
                configuration,
                section,
            )
            for task in tasks:
                account_id = task.get("account_id")
                region = task.get("region")
                if shares_by_region_portfolio_account.get(region) is None:
                    shares_by_region_portfolio_account[region] = {}
                if shares_by_region_portfolio_account[region].get(portfolio) is None:
                    shares_by_region_portfolio_account[region][portfolio] = {}
                result = self.get_account(account_id)
                result[section] = launch_details
                shares_by_region_portfolio_account[region][portfolio][
                    account_id
                ] = result
        return shares_by_region_portfolio_account

    def get_accounts_by_region(self):
        accounts_by_region = {}
        for account in self.get("accounts"):
            account_regions = list()
            account_regions += account.get("enabled", [])
            account_regions += account.get("regions_enabled", [])
            account_regions += account.get("enabled_regions", [])
            account_regions.append(account.get("default_region"))
            for r in account_regions:
                if accounts_by_region.get(r) is None:
                    accounts_by_region[r] = []
                accounts_by_region[r].append(account)
        return accounts_by_region

    def get_task_defs_from_details(
        self,
        puppet_account_id,
        launch_name,
        configuration,
        launch_or_spoke_local_portfolio,
    ):
        launch_details = self.get(launch_or_spoke_local_portfolio).get(launch_name)
        accounts = self.get("accounts")
        if launch_details is None:
            raise Exception(f"launch_details is None for {launch_name}")
        if launch_or_spoke_local_portfolio == "lambda-invocations":
            deploy_to = launch_details.get("invoke_for")
        if launch_or_spoke_local_portfolio == constants.SERVICE_CONTROL_POLICIES:
            deploy_to = launch_details.get("apply_to")
        elif launch_or_spoke_local_portfolio == "launches":
            deploy_to = launch_details.get("deploy_to")
        elif launch_or_spoke_local_portfolio == "spoke-local-portfolios":
            deploy_to = launch_details.get("deploy_to") or launch_details.get(
                "share_with"
            )
        elif launch_or_spoke_local_portfolio == "imported-portfolios":
            deploy_to = launch_details.get("deploy_to") or launch_details.get(
                "share_with"
            )
        task_defs = []
        for tag_list_item in deploy_to.get("tags", []):
            for account in accounts:
                for tag in account.get("tags", []):
                    if tag == tag_list_item.get("tag"):
                        tag_account_def = deepcopy(configuration)
                        tag_account_def["account_id"] = account.get("account_id")
                        tag_account_def["account_parameters"] = account.get(
                            "parameters", {}
                        )

                        regions = tag_list_item.get("regions", "default_region")
                        if isinstance(regions, str):
                            if regions in [
                                "enabled",
                                "regions_enabled",
                                "enabled_regions",
                            ]:
                                for region_enabled in account.get("regions_enabled"):
                                    region_tag_account_def = deepcopy(tag_account_def)
                                    region_tag_account_def["region"] = region_enabled
                                    task_defs.append(region_tag_account_def)
                            elif regions == "default_region":
                                region_tag_account_def = deepcopy(tag_account_def)
                                region_tag_account_def["region"] = account.get(
                                    "default_region"
                                )
                                task_defs.append(region_tag_account_def)
                            elif regions == "all":
                                all_regions = regions_configured
                                for region_enabled in all_regions:
                                    region_tag_account_def = deepcopy(tag_account_def)
                                    region_tag_account_def["region"] = region_enabled
                                    task_defs.append(region_tag_account_def)
                            else:
                                raise Exception(
                                    f"Unsupported regions {regions} setting for launch: {launch_name}"
                                )
                        elif isinstance(regions, list):
                            for region in regions:
                                region_tag_account_def = deepcopy(tag_account_def)
                                region_tag_account_def["region"] = region
                                task_defs.append(region_tag_account_def)
                        elif isinstance(regions, tuple):
                            for region in regions:
                                region_tag_account_def = deepcopy(tag_account_def)
                                region_tag_account_def["region"] = region
                                task_defs.append(region_tag_account_def)
                        else:
                            raise Exception(
                                f"Unexpected regions of {regions} set for launch {launch_name}"
                            )

        for account_list_item in deploy_to.get("accounts", []):
            for account in accounts:
                if account.get("account_id") == account_list_item.get("account_id"):
                    account_account_def = deepcopy(configuration)
                    account_account_def["account_id"] = account.get("account_id")
                    account_account_def["account_parameters"] = account.get(
                        "parameters", {}
                    )

                    regions = account_list_item.get("regions", "default_region")
                    if isinstance(regions, str):
                        if regions in ["enabled", "regions_enabled", "enabled_regions"]:
                            for region_enabled in account.get("regions_enabled"):
                                region_account_account_def = deepcopy(
                                    account_account_def
                                )
                                region_account_account_def["region"] = region_enabled
                                task_defs.append(region_account_account_def)
                        elif regions in ["default_region"]:
                            region_account_account_def = deepcopy(account_account_def)
                            region_account_account_def["region"] = account.get(
                                "default_region"
                            )
                            task_defs.append(region_account_account_def)
                        elif regions in ["all"]:
                            all_regions = regions_configured
                            for region_enabled in all_regions:
                                region_account_account_def = deepcopy(
                                    account_account_def
                                )
                                region_account_account_def["region"] = region_enabled
                                task_defs.append(region_account_account_def)
                        else:
                            raise Exception(
                                f"Unsupported regions {regions} setting for launch: {launch_name}"
                            )

                    elif isinstance(regions, list):
                        for region in regions:
                            region_account_account_def = deepcopy(account_account_def)
                            region_account_account_def["region"] = region
                            task_defs.append(region_account_account_def)
                    elif isinstance(regions, tuple):
                        for region in regions:
                            region_account_account_def = deepcopy(account_account_def)
                            region_account_account_def["region"] = region
                            task_defs.append(region_account_account_def)
                    else:
                        raise Exception(
                            f"Unexpected regions of {regions} set for launch {launch_name}"
                        )
        return task_defs


def create_minimal_manifest(manifest):
    minimal_manifest = deepcopy(manifest)
    # minimal_manifest[constants.ACCOUNTS] = dict()
    minimal_manifest[constants.LAUNCHES] = dict()
    minimal_manifest[constants.STACKS] = dict()
    minimal_manifest[constants.SPOKE_LOCAL_PORTFOLIOS] = dict()
    minimal_manifest[constants.LAMBDA_INVOCATIONS] = dict()
    minimal_manifest[constants.ASSERTIONS] = dict()
    minimal_manifest[constants.SERVICE_CONTROL_POLICIES] = dict()
    minimal_manifest[constants.TAG_POLICIES] = dict()
    minimal_manifest[constants.APPS] = dict()
    minimal_manifest[constants.WORKSPACES] = dict()
    return minimal_manifest


def convert_to_graph(expanded_manifest, G):
    for section in constants.ALL_SECTION_NAMES:
        for item_name, item_details in expanded_manifest.get(section, {}).items():
            uid = f"{section}|{item_name}"
            data = dict(
                section=section,
                item_name=item_name,
            )
            data.update(item_details)
            G.add_nodes_from(
                [
                    (uid, data),
                ]
            )

    for section in constants.ALL_SECTION_NAMES:
        for item_name, item_details in expanded_manifest.get(section, {}).items():
            uid = f"{section}|{item_name}"
            for d in item_details.get("depends_on", []):
                if isinstance(d, str):
                    G.add_edge(uid, f"{constants.LAUNCHES}|{d}")
                else:
                    G.add_edge(
                        uid,
                        f"{constants.SECTION_SINGULAR_TO_PLURAL.get(d.get('type'))}|{d.get('name')}",
                    )
    return G


def explode(expanded_manifest):
    G = convert_to_graph(expanded_manifest, nx.Graph())

    S = [G.subgraph(c).copy() for c in nx.connected_components(G)]
    exploded = list()
    for s in S:
        m = create_minimal_manifest(expanded_manifest)
        for node in s.nodes(data=True):
            data = deepcopy(node[1])
            del data["section"]
            del data["item_name"]
            m[node[1].get("section")][node[1].get("item_name")] = data
        exploded.append(m)
    return exploded


def isolate(expanded_manifest, subset):
    section = subset["single_action_section"]
    item = subset["single_action_item"]
    uid = f"{section}|{item}"

    m = create_minimal_manifest(expanded_manifest)
    G = convert_to_graph(expanded_manifest, nx.DiGraph())
    node = G.nodes.get(uid)
    data = deepcopy(node)
    del data["section"]
    del data["item_name"]
    m[node.get("section")][node.get("item_name")] = data

    if subset.get("single_action_include_dependencies", False):
        click.echo("Including dependencies")
        for dependency in nx.edge_dfs(G, uid, orientation="original"):
            link, dependency_name, direction = dependency
            node = G.nodes.get(dependency_name)
            data = deepcopy(node)
            del data["section"]
            del data["item_name"]
            m[node.get("section")][node.get("item_name")] = data

    if subset.get("single_action_include_reverse_dependencies", False):
        click.echo("Including reverse dependencies")
        for reverse_dependency in nx.edge_dfs(G, uid, orientation="reverse"):
            reverse_dependency_name, link, direction = reverse_dependency
            node = G.nodes.get(reverse_dependency_name)
            data = deepcopy(node)
            del data["section"]
            del data["item_name"]
            m[node.get("section")][node.get("item_name")] = data

            for dependency in nx.edge_dfs(
                G, reverse_dependency_name, orientation="original"
            ):
                link, dependency_name, direction = dependency
                node = G.nodes.get(dependency_name)
                data = deepcopy(node)
                del data["section"]
                del data["item_name"]
                m[node.get("section")][node.get("item_name")] = data

    return m


def parse_conditions(manifest):
    for section_name in constants.SECTIONS_THAT_SUPPORT_CONDITIONS:
        for item_name, item in manifest.get(section_name, {}).items():
            if item.get("condition"):
                if isinstance(item.get("condition"), str):
                    condition = manifest.get("conditions").get(item.get("condition"))
                    if not condition.get_result():
                        manifest[section_name][item_name][
                            constants.MANIFEST_STATUS_FIELD_NAME
                        ] = constants.MANIFEST_STATUS_FIELD_VALUE_IGNORED
                        logger.info(
                            f"Removed {item_name} from {section_name} because condition ({item.get('condition')}) evaluated to false"
                        )
                else:
                    should_ignore = True
                    for condition_to_check in item.get("condition"):
                        condition = manifest.get("conditions").get(condition_to_check)
                        if condition.get_result():
                            should_ignore = False
                    if should_ignore:
                        manifest[section_name][item_name][
                            constants.MANIFEST_STATUS_FIELD_NAME
                        ] = constants.MANIFEST_STATUS_FIELD_VALUE_IGNORED
                        logger.info(
                            f"Removed {item_name} from {section_name} because condition ({item.get('condition')}) evaluated to false"
                        )
    return manifest


def rewrite_c7n_cloudtrails(manifest):
    if manifest.get(constants.C7N_AWS_CLOUDTRAILS):
        manifest[constants.C7N_AWS_LAMBDAS] = manifest[constants.C7N_AWS_CLOUDTRAILS]
        del manifest[constants.C7N_AWS_CLOUDTRAILS]
        for item_name, item in manifest[constants.C7N_AWS_LAMBDAS].items():
            for policy in item["policies"]:
                policy["mode"]["type"] = "cloudtrail"
    return manifest


def rewrite_organizational_units(manifest, conversions, client):
    for item_name, item in manifest.get(constants.ORGANIZATIONAL_UNITS, {}).items():
        path = item.get("path")
        if not item.get("parent_ou_id"):
            parent_path = os.path.dirname(path)
            if conversions.get(parent_path):
                item["parent_ou_id"] = conversions.get(parent_path)
        if not item.get("name"):
            item["name"] = os.path.basename(path)

    return manifest



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
from servicecatalog_puppet import config, constants, task_reference_constants
from servicecatalog_puppet.commands.task_reference_helpers.generators import portfolios


def get_imported_portfolio_common_args(
    task_to_add, all_tasks_task_reference, extra_dependencies_by_reference=[]
):
    return dict(
        status=task_to_add.get("status"),
        account_id=task_to_add.get("account_id"),
        region=task_to_add.get("region"),
        portfolio=task_to_add.get("portfolio"),
        execution=task_to_add.get("execution"),
        dependencies_by_reference=[all_tasks_task_reference]
        + extra_dependencies_by_reference,
        portfolio_task_reference=all_tasks_task_reference,
    )


def add_puppet_associations_for_when_not_sharing_with_puppet_account(
    all_tasks,
    all_tasks_task_reference,
    puppet_account_id,
    task_to_add,
    hub_portfolio_ref,
    dependencies_for_import_portfolios_sub_tasks,
):
    # GET THE HUB PORTFOLIO
    if not all_tasks.get(hub_portfolio_ref):
        all_tasks[hub_portfolio_ref] = {
            "puppet_account_id": puppet_account_id,
            "task_reference": hub_portfolio_ref,
            "dependencies_by_reference": [],
            "account_id": puppet_account_id,
            "region": task_to_add.get("region"),
            "portfolio": task_to_add.get("portfolio"),
            "status": task_to_add.get("status"),
            "execution": task_to_add.get("execution"),
            "section_name": constants.PORTFOLIO_LOCAL,
            task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
            task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
            task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
        }
    all_tasks[hub_portfolio_ref][
        task_reference_constants.MANIFEST_SECTION_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
    all_tasks[hub_portfolio_ref][task_reference_constants.MANIFEST_ITEM_NAMES].update(
        task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES)
    )
    all_tasks[hub_portfolio_ref][task_reference_constants.MANIFEST_ACCOUNT_IDS].update(
        task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS)
    )
    del all_tasks[all_tasks_task_reference]
    # all_tasks[all_tasks_task_reference]["portfolio_task_reference"] = hub_portfolio_ref
    # all_tasks[all_tasks_task_reference]["dependencies_by_reference"].append(
    #     hub_portfolio_ref
    # )

    # CREATE THE HUB ASSOCIATIONS
    hub_portfolio_puppet_association_ref = f"{constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION}-{puppet_account_id}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
    if not all_tasks.get(hub_portfolio_puppet_association_ref):
        all_tasks[hub_portfolio_puppet_association_ref] = {
            "puppet_account_id": puppet_account_id,
            "task_reference": hub_portfolio_puppet_association_ref,
            "portfolio_task_reference": hub_portfolio_ref,
            "dependencies_by_reference": [
                hub_portfolio_ref,
                constants.CREATE_POLICIES,
            ],
            "account_id": puppet_account_id,
            "region": task_to_add.get("region"),
            "portfolio": task_to_add.get("portfolio"),
            "execution": task_to_add.get("execution"),
            "section_name": constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION,
            task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
            task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
            task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
        }
    all_tasks[hub_portfolio_puppet_association_ref][
        task_reference_constants.MANIFEST_SECTION_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
    all_tasks[hub_portfolio_puppet_association_ref][
        task_reference_constants.MANIFEST_ITEM_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
    all_tasks[hub_portfolio_puppet_association_ref][
        task_reference_constants.MANIFEST_ACCOUNT_IDS
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

    # CREATE SPOKE ASSOCIATIONS
    spoke_portfolio_puppet_association_ref = f"{constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION}-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
    if not all_tasks.get(spoke_portfolio_puppet_association_ref):
        all_tasks[spoke_portfolio_puppet_association_ref] = {
            "puppet_account_id": puppet_account_id,
            "task_reference": spoke_portfolio_puppet_association_ref,
            "portfolio_task_reference": hub_portfolio_ref,
            "dependencies_by_reference": [
                hub_portfolio_ref,
                constants.CREATE_POLICIES,
            ]
            + dependencies_for_import_portfolios_sub_tasks,
            "account_id": task_to_add.get("account_id"),
            "region": task_to_add.get("region"),
            "portfolio": task_to_add.get("portfolio"),
            "execution": task_to_add.get("execution"),
            "section_name": constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION,
            task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
            task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
            task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
        }
    all_tasks[spoke_portfolio_puppet_association_ref][
        task_reference_constants.MANIFEST_SECTION_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
    all_tasks[spoke_portfolio_puppet_association_ref][
        task_reference_constants.MANIFEST_ITEM_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
    all_tasks[spoke_portfolio_puppet_association_ref][
        task_reference_constants.MANIFEST_ACCOUNT_IDS
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

    return hub_portfolio_ref


def setup_sharing_for_portfolio(
    all_tasks,
    all_tasks_task_reference,
    item_name,
    puppet_account_id,
    section_name,
    task_reference,
    task_to_add,
    hub_portfolio_ref,
):
    # SHARE THE PORTFOLIO
    sharing_mode = task_to_add.get(
        "sharing_mode", config.get_global_sharing_mode_default()
    )
    if sharing_mode == constants.SHARING_MODE_ACCOUNT:
        share_and_accept_ref = f"portfolio_share_and_accept-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
        sharing_type = "ACCOUNT"
    elif sharing_mode == constants.SHARING_MODE_AWS_ORGANIZATIONS:
        share_and_accept_ref = f"portfolio_share_and_accept-{task_to_add.get('ou')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
        sharing_type = (
            "ORGANIZATIONAL_UNIT"
            if task_to_add.get("ou")[0:3] == "ou-"
            else "ORGANIZATION"
        )
    else:
        raise Exception(f"Unknown sharing mode: {sharing_mode}")

    describe_portfolio_shares_task_ref = (
        portfolios.get_or_create_describe_portfolio_shares_task_ref(
            all_tasks,
            puppet_account_id,
            sharing_type,
            hub_portfolio_ref,
            task_to_add,
        )
    )

    if not all_tasks.get(share_and_accept_ref):
        all_tasks[share_and_accept_ref] = {
            "puppet_account_id": puppet_account_id,
            "account_id": task_to_add.get("account_id"),
            "region": task_to_add.get("region"),
            "task_reference": share_and_accept_ref,
            "share_tag_options": task_to_add.get("share_tag_options"),
            "share_principals": task_to_add.get("share_principals"),
            "dependencies_by_reference": [
                hub_portfolio_ref,
                constants.CREATE_POLICIES,
                describe_portfolio_shares_task_ref,
            ],
            "describe_portfolio_shares_task_ref": describe_portfolio_shares_task_ref,
            "portfolio": task_to_add.get("portfolio"),
            "execution": task_to_add.get("execution"),
            "portfolio_task_reference": hub_portfolio_ref,
            "section_name": f"portfolio-share-and-accept-{sharing_mode.lower()}",
            "ou_to_share_with": task_to_add.get("ou"),
            task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
            task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
            task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
        }
    all_tasks[share_and_accept_ref][
        task_reference_constants.MANIFEST_SECTION_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
    all_tasks[share_and_accept_ref][
        task_reference_constants.MANIFEST_ITEM_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
    all_tasks[share_and_accept_ref][
        task_reference_constants.MANIFEST_ACCOUNT_IDS
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))
    return share_and_accept_ref


def handle_imported_portfolios(
    all_tasks,
    all_tasks_task_reference,
    item_name,
    puppet_account_id,
    section_name,
    task_reference,
    task_to_add,
):
    is_sharing_with_puppet_account = task_to_add.get("account_id") == puppet_account_id
    #
    if task_to_add.get("status") == constants.TERMINATED:
        # DELETE THE ASSOCIATION IF IT EXISTS
        if task_to_add.get("associations"):
            shared_ref = f"{section_name}-{item_name}-{task_to_add.get('account_id')}-{task_to_add.get('region')}"
            ref = f"portfolio_associations-{shared_ref}"
            if not all_tasks.get(ref):
                all_tasks[ref] = {
                    "status": task_to_add.get("status"),
                    "account_id": task_to_add.get("account_id"),
                    "region": task_to_add.get("region"),
                    "portfolio": task_to_add.get("portfolio"),
                    "execution": task_to_add.get("execution"),
                    "dependencies_by_reference": [constants.CREATE_POLICIES],
                    "task_reference": ref,
                    "spoke_local_portfolio_name": item_name,
                    "section_name": constants.PORTFOLIO_ASSOCIATIONS,
                    "associations": task_to_add.get("associations"),
                    task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                    task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                    task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
                }
            all_tasks[ref][task_reference_constants.MANIFEST_SECTION_NAMES].update(
                task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES)
            )
            all_tasks[ref][task_reference_constants.MANIFEST_ITEM_NAMES].update(
                task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES)
            )
            all_tasks[ref][task_reference_constants.MANIFEST_ACCOUNT_IDS].update(
                task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS)
            )
            del all_tasks[
                all_tasks_task_reference
            ]  # DO NOT NEED THIS TASK TO DO ANYTHING AT THE MOMENT
    else:
        dependencies_for_import_portfolios_sub_tasks = list()
        if is_sharing_with_puppet_account:
            target_portfolio_ref = (
                add_puppet_associations_for_when_sharing_with_puppet_account(
                    all_tasks, all_tasks_task_reference, puppet_account_id, task_to_add
                )
            )

        else:
            hub_portfolio_ref = f"{constants.PORTFOLIO_LOCAL}-{puppet_account_id}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
            share_and_accept_ref = setup_sharing_for_portfolio(
                all_tasks,
                all_tasks_task_reference,
                item_name,
                puppet_account_id,
                section_name,
                task_reference,
                task_to_add,
                hub_portfolio_ref,
            )
            dependencies_for_import_portfolios_sub_tasks.append(share_and_accept_ref)
            target_portfolio_ref = (
                add_puppet_associations_for_when_not_sharing_with_puppet_account(
                    all_tasks,
                    all_tasks_task_reference,
                    puppet_account_id,
                    task_to_add,
                    hub_portfolio_ref,
                    dependencies_for_import_portfolios_sub_tasks,
                )
            )

        # need to add the sharing tasks to the dependencies
        if task_to_add.get("associations"):
            shared_ref = f"{section_name}-{item_name}-{task_to_add.get('account_id')}-{task_to_add.get('region')}"
            ref = f"portfolio_associations-{shared_ref}"
            all_tasks[ref] = dict(
                **get_imported_portfolio_common_args(
                    task_to_add,
                    target_portfolio_ref,
                    [constants.CREATE_POLICIES]
                    + dependencies_for_import_portfolios_sub_tasks,
                ),
                task_reference=ref,
                spoke_local_portfolio_name=item_name,
                section_name=constants.PORTFOLIO_ASSOCIATIONS,
                associations=task_to_add.get("associations"),
            )
            all_tasks[ref][task_reference_constants.MANIFEST_SECTION_NAMES] = dict(
                **task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES)
            )
            all_tasks[ref][task_reference_constants.MANIFEST_ITEM_NAMES] = dict(
                **task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES)
            )
            all_tasks[ref][task_reference_constants.MANIFEST_ACCOUNT_IDS] = dict(
                **task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS)
            )

    if task_to_add.get("was_a_spoke_local_portfolio"):
        # DELETE THE LAUNCH CONSTRAINTS IF IT EXISTS
        shared_ref = f"{section_name}-{item_name}-{task_to_add.get('account_id')}-{task_to_add.get('region')}"
        ref = f"launch_constraints-{shared_ref}"
        if not all_tasks.get(ref):
            all_tasks[ref] = {
                "status": constants.TERMINATED,
                "account_id": task_to_add.get("account_id"),
                "region": task_to_add.get("region"),
                "portfolio": task_to_add.get("portfolio"),
                "execution": task_to_add.get("execution"),
                "dependencies_by_reference": [constants.CREATE_POLICIES]
                + dependencies_for_import_portfolios_sub_tasks,
                "task_reference": ref,
                "section_name": constants.PORTFOLIO_CONSTRAINTS_LAUNCH,
                "spoke_local_portfolio_name": item_name,
                "launch_constraints": task_to_add.get("launch_constraints"),
                task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
            }
        all_tasks[ref][task_reference_constants.MANIFEST_SECTION_NAMES].update(
            task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES)
        )
        all_tasks[ref][task_reference_constants.MANIFEST_ITEM_NAMES].update(
            task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES)
        )
        all_tasks[ref][task_reference_constants.MANIFEST_ACCOUNT_IDS].update(
            task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS)
        )
        # DELETE THE RESOURCE UPDATE CONSTRAINTS IF IT EXISTS
        shared_ref = f"{section_name}-{item_name}-{task_to_add.get('account_id')}-{task_to_add.get('region')}"
        ref = f"resource_update_constraints-{shared_ref}"
        if not all_tasks.get(ref):
            all_tasks[ref] = {
                "status": constants.TERMINATED,
                "account_id": task_to_add.get("account_id"),
                "region": task_to_add.get("region"),
                "portfolio": task_to_add.get("portfolio"),
                "execution": task_to_add.get("execution"),
                "dependencies_by_reference": [constants.CREATE_POLICIES]
                + dependencies_for_import_portfolios_sub_tasks,
                "task_reference": ref,
                "section_name": constants.PORTFOLIO_CONSTRAINTS_RESOURCE_UPDATE,
                "resource_update_constraints": task_to_add.get(
                    "resource_update_constraints"
                ),
                "spoke_local_portfolio_name": item_name,
                task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
            }
        all_tasks[ref][task_reference_constants.MANIFEST_SECTION_NAMES].update(
            task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES)
        )
        all_tasks[ref][task_reference_constants.MANIFEST_ITEM_NAMES].update(
            task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES)
        )
        all_tasks[ref][task_reference_constants.MANIFEST_ACCOUNT_IDS].update(
            task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS)
        )


def add_puppet_associations_for_when_sharing_with_puppet_account(
    all_tasks, all_tasks_task_reference, puppet_account_id, task_to_add
):
    # CREATE SPOKE ASSOCIATIONS
    spoke_portfolio_puppet_association_ref = f"{constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION}-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
    if not all_tasks.get(spoke_portfolio_puppet_association_ref):
        all_tasks[spoke_portfolio_puppet_association_ref] = {
            "puppet_account_id": puppet_account_id,
            "task_reference": spoke_portfolio_puppet_association_ref,
            "portfolio_task_reference": all_tasks_task_reference,
            "dependencies_by_reference": [
                all_tasks_task_reference,
                constants.CREATE_POLICIES,
            ],
            "account_id": task_to_add.get("account_id"),
            "region": task_to_add.get("region"),
            "portfolio": task_to_add.get("portfolio"),
            "execution": task_to_add.get("execution"),
            "section_name": constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION,
            task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
            task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
            task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
        }
    all_tasks[spoke_portfolio_puppet_association_ref][
        task_reference_constants.MANIFEST_SECTION_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
    all_tasks[spoke_portfolio_puppet_association_ref][
        task_reference_constants.MANIFEST_ITEM_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
    all_tasks[spoke_portfolio_puppet_association_ref][
        task_reference_constants.MANIFEST_ACCOUNT_IDS
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))
    return spoke_portfolio_puppet_association_ref



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import re

import luigi
import troposphere as t
from troposphere import servicecatalog

from servicecatalog_puppet import config, constants, utils
from servicecatalog_puppet.workflow.dependencies import tasks


class CreateLaunchRoleConstraintsForSpokeLocalPortfolioTask(tasks.TaskWithReference):
    portfolio_task_reference = luigi.Parameter()
    spoke_local_portfolio_name = luigi.Parameter()
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    launch_constraints = luigi.ListParameter()

    portfolio_get_all_products_and_their_versions_ref = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "spoke_local_portfolio_name": self.spoke_local_portfolio_name,
            "portfolio": self.portfolio,
            "region": self.region,
            "account_id": self.account_id,
        }

    def run(self):
        portfolio_details = self.get_output_from_reference_dependency(
            self.portfolio_task_reference
        )
        portfolio_id = portfolio_details.get("Id")
        products_and_their_versions = self.get_output_from_reference_dependency(
            self.portfolio_get_all_products_and_their_versions_ref
        )

        tpl = t.Template()
        tpl.description = f"launch constraints for {self.portfolio}"

        for product_to_constrain, roles in self.launch_constraints.get(
            "products", {}
        ).items():
            for role_arn in roles:
                for (
                    product_name,
                    product_details,
                ) in products_and_their_versions.items():
                    if re.match(
                        product_to_constrain,
                        product_name,
                    ):
                        product_id = product_details.get("ProductId")
                        tpl.add_resource(
                            servicecatalog.LaunchRoleConstraint(
                                f"L{portfolio_id}B{product_id}C{role_arn.split(':')[-1]}".replace(
                                    "-", ""
                                ).replace(
                                    "/", ""
                                ),
                                PortfolioId=portfolio_id,
                                ProductId=product_id,
                                RoleArn=role_arn.replace(
                                    "${AWS::AccountId}", self.account_id
                                ),
                            )
                        )

        template_to_use = tpl.to_yaml()

        with self.spoke_regional_client("cloudformation") as cloudformation:
            stack_name = f"launch-constraints-for-{utils.slugify_for_cloudformation_stack_name(self.portfolio)}"
            cloudformation.create_or_update(
                StackName=stack_name,
                TemplateBody=template_to_use,
                NotificationARNs=(
                    [
                        f"arn:{config.get_partition()}:sns:{self.region}:{self.puppet_account_id}:servicecatalog-puppet-cloudformation-regional-events"
                    ]
                    if self.should_use_sns
                    else []
                ),
                ShouldDeleteRollbackComplete=self.should_delete_rollback_complete_stacks,
                Tags=self.initialiser_stack_tags,
            )
            self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import re

import luigi
import troposphere as t
from troposphere import servicecatalog

from servicecatalog_puppet import config, constants, utils
from servicecatalog_puppet.workflow.dependencies import tasks
from servicecatalog_puppet.workflow.portfolio.associations import association_utils


class CreateAssociationsForSpokeLocalPortfolioTask(tasks.TaskWithReference):
    portfolio_task_reference = luigi.Parameter()
    spoke_local_portfolio_name = luigi.Parameter()
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio = luigi.Parameter()

    associations = luigi.ListParameter(default=[])
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "spoke_local_portfolio_name": self.spoke_local_portfolio_name,
            "portfolio": self.portfolio,
            "region": self.region,
            "account_id": self.account_id,
        }

    def run(self):
        portfolio_details = self.get_output_from_reference_dependency(
            self.portfolio_task_reference
        )
        portfolio_id = portfolio_details.get("Id")

        haystack = "".join(self.associations)
        used_wildcard = "*" in haystack
        used_roles = ":role/" in haystack
        used_users = ":user/" in haystack
        used_groups = ":group/" in haystack

        if used_wildcard:
            roles = list()
            users = list()
            groups = list()
            with self.spoke_regional_client("iam") as iam:
                if used_roles:
                    paginator = iam.get_paginator("list_roles")
                    for page in paginator.paginate():
                        roles += page.get("Roles", [])
                if used_users:
                    paginator = iam.get_paginator("list_users")
                    for page in paginator.paginate():
                        users += page.get("Users", [])
                if used_groups:
                    paginator = iam.get_paginator("list_groups")
                    for page in paginator.paginate():
                        groups += page.get("Groups", [])

            associations_to_use = list()
            for association in self.associations:
                if "*" in association:
                    association_as_a_regex = re.escape(
                        association.replace("${AWS::AccountId}", self.account_id)
                    ).replace("\\*", "(.*)")
                    if ":role/" in association:
                        for role_object in roles:
                            role = role_object.get("Arn")
                            if re.match(association_as_a_regex, role):
                                associations_to_use.append(role)
                    if ":user/" in association:
                        for user_object in users:
                            user = user_object.get("Arn")
                            if re.match(association_as_a_regex, user):
                                associations_to_use.append(user)
                    if ":group/" in association:
                        for group_object in groups:
                            group = group_object.get("Arn")
                            if re.match(association_as_a_regex, group):
                                associations_to_use.append(group)
                else:
                    associations_to_use.append(association)
        else:
            associations_to_use = self.associations

        with self.spoke_regional_client("cloudformation") as cloudformation:
            stack_name = (
                association_utils.generate_stack_name_for_associations_by_item_name(
                    self.portfolio
                )
            )
            if associations_to_use:
                tpl = t.Template()
                tpl.description = f"Associations for {self.portfolio}"
                for association_to_use in associations_to_use:
                    logical_name = "".join(
                        filter(
                            str.isalnum,
                            f"{portfolio_id}{association_to_use.split(':')[-1]}",
                        )
                    )
                    tpl.add_resource(
                        servicecatalog.PortfolioPrincipalAssociation(
                            logical_name,
                            PortfolioId=portfolio_id,
                            PrincipalARN=t.Sub(association_to_use),
                            PrincipalType="IAM",
                        )
                    )
                template = tpl.to_yaml()
                cloudformation.create_or_update(
                    StackName=stack_name,
                    TemplateBody=template,
                    NotificationARNs=(
                        [
                            f"arn:{config.get_partition()}:sns:{self.region}:{self.puppet_account_id}:servicecatalog-puppet-cloudformation-regional-events"
                        ]
                        if self.should_use_sns
                        else []
                    ),
                    ShouldDeleteRollbackComplete=self.should_delete_rollback_complete_stacks,
                    Tags=self.initialiser_stack_tags,
                )
            else:
                cloudformation.ensure_deleted(StackName=stack_name)
        self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import time

import luigi

from servicecatalog_puppet import config, constants
from servicecatalog_puppet.workflow.dependencies import tasks


class TerminateAssociationTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    portfolio_task_reference = luigi.Parameter()

    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "region": self.region,
            "account_id": self.account_id,
        }

    def run(self):
        portfolio_details = self.get_output_from_reference_dependency(
            self.portfolio_task_reference
        )
        portfolio_id = portfolio_details.get("Id")
        if portfolio_id is None:
            self.write_empty_output()
        else:
            with self.spoke_regional_client("servicecatalog") as servicecatalog:
                principal_to_associate = config.get_puppet_role_arn(self.account_id)
                was_present = self.check_if_present(
                    portfolio_id, principal_to_associate, servicecatalog
                )
                if was_present:
                    servicecatalog.disassociate_principal_from_portfolio(
                        PortfolioId=portfolio_id,
                        PrincipalARN=principal_to_associate,
                    )
                is_still_present = True
                while is_still_present:
                    time.sleep(1)
                    is_still_present = self.check_if_present(
                        portfolio_id, principal_to_associate, servicecatalog
                    )

            self.write_empty_output()

    def check_if_present(self, portfolio_id, principal_to_associate, servicecatalog):
        paginator = servicecatalog.get_paginator("list_principals_for_portfolio")
        for page in paginator.paginate(PortfolioId=portfolio_id):
            for principal in page.get("Principals", []):
                if principal_to_associate == principal.get("PrincipalARN"):
                    return True
        return False



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import json
from copy import deepcopy

import jmespath
import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


class GetS3ObjectTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    key = luigi.Parameter()
    region = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
            "account_id": self.account_id,
            "region": self.region,
            "key": self.key,
        }

    def run(self):
        with self.spoke_regional_client("s3") as s3:
            object = (
                s3.get_object(
                    Bucket=f"sc-puppet-parameters-{self.account_id}", Key=self.key
                )
                .get("Body")
                .read()
            )
            self.write_output(object, skip_json_encode=True)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import config, constants
from servicecatalog_puppet.workflow.dependencies import tasks


class GeneratePolicies(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    region = luigi.Parameter()

    organizations_to_share_with = luigi.ListParameter()
    ous_to_share_with = luigi.ListParameter()
    accounts_to_share_with = luigi.ListParameter()
    cachable_level = constants.CACHE_LEVEL_NO_CACHE

    def params_for_results_display(self):
        return {
            "account_id": self.account_id,
            "region": self.region,
        }

    def run(self):
        sharing_policies = dict(
            accounts=self.accounts_to_share_with,
            ous=self.ous_to_share_with,
            organizations=self.organizations_to_share_with,
        )

        if len(sharing_policies.get("accounts", [])) > 50:
            self.warning(
                "You have specified more than 50 accounts will not create the eventbus policy and spoke execution mode will not work"
            )
        template = config.env.get_template("policies.template.yaml.j2").render(
            sharing_policies=sharing_policies,
            VERSION=constants.VERSION,
            HOME_REGION=constants.HOME_REGION,
        )

        with self.spoke_regional_client("cloudformation") as cloudformation:
            cloudformation.create_or_update(
                ShouldUseChangeSets=False,
                StackName="servicecatalog-puppet-policies",
                TemplateBody=template,
                NotificationARNs=(
                    [
                        f"arn:{config.get_partition()}:sns:{self.region}:{self.puppet_account_id}:servicecatalog-puppet-cloudformation-regional-events"
                    ]
                    if self.should_use_sns
                    else []
                ),
                ShouldDeleteRollbackComplete=self.should_delete_rollback_complete_stacks,
                Tags=self.initialiser_stack_tags,
            )

        self.write_empty_output()
